import os
import json
import logging
from itertools import chain
from typing import Literal

import pandas as pd
import numpy as np

from neuroglancer.coordinate_space import CoordinateSpace
from neuroglancer.viewer_state import AnnotationPropertySpec

from ..util import annotation_property_specs
from ._util import _encode_uint64_series, _geometry_cols, TableHandle
from ._id import _write_annotations_by_id
from ._relationships import _write_annotations_by_relationships, _encode_relationships
from ._spatial import _write_annotations_by_spatial_chunk

logger = logging.getLogger(__name__)


def write_precomputed_annotations(
    df: pd.DataFrame | TableHandle,
    coord_space: CoordinateSpace | str | list[str] | dict[str, list],
    annotation_type: Literal['point', 'line', 'ellipsoid', 'axis_aligned_bounding_box'],
    properties: list[str] | list[AnnotationPropertySpec] | dict[str, AnnotationPropertySpec] | list[dict] = (),
    relationships: list[str] = (),
    output_dir: str = 'annotations',
    write_sharded: bool = True,
    *,
    write_by_id: bool = True,
    write_by_relationship: bool = True,
    write_by_spatial_chunk: bool = True,
    num_spatial_levels: int = 7,
    target_chunk_limit: int = 10_000,
    shuffle_before_assigning_spatial_levels: bool = True,
    description: str = "",
):
    """
    Export the data from a pandas DataFrame into neuroglancer's precomputed annotations format
    as described in the `neuroglancer spec <https://github.com/google/neuroglancer/blob/master/src/datasource/precomputed/annotations.md>`_.

    A progress bar is shown when writing each portion of the export (annotation ID index, related ID indexes),
    but there may be a significant amount of preprocessing time that occurs before the actual writing begins.

    Note:

        Internally, the data will be copied during processing and again 
        during writing, incurring significant RAM usage for large datasets.
        To save at least some RAM, you can wrap your dataframe in a TableHandle
        and then delete your own reference to the dataframe before calling this function.
        The TableHandle's reference will be deleted internally as soon as possible
        (after the data is transformed for writing, before this function returns).

    Args:
        df:
            DataFrame or TableHandle.
            The index of the DataFrame is used as the annotation ID, so it must be unique.
            The required columns depend on the annotation_type and the coordinate space.
            For example, assuming ``coord_space.names == ['x', 'y', 'z']``,
            then provide the following columns:

            - For point annotations, provide ['x', 'y', 'z']
            - For line annotations or axis_aligned_bounding_box annotations,
              provide ['xa', 'ya', 'za', 'xb', 'yb', 'zb']
            - For ellipsoid annotations, provide ['x', 'y', 'z', 'rx', 'ry', 'rz']
              for the center point and radii.

            You may also provide additional columns to use as annotation properties, in which
            case their column names should be listed in the 'properties' argument. (See below.)

            If you provide a TableHandle, the handle's reference will be unset before this
            function returns, deleting your data if you didn't retain a reference to it yourself.
            (If you do retain a reference, it defeats the point of using a TableHandle in the first place.)

        coord_space:
            ``neuroglancer.coordinate_space.CoordinateSpace`` or equivalent.
            The coordinate space of the annotations.
            Among other things, this determines which input columns represent the annotation geometry.
            For convenience, we accept a couple different formats for the coordinate space,
            assuming a default scale of 1 nm if no scale/units are provided.

            Examples (all equivalent):

            .. code-block:: python

                >>> coord_space = "xyz"
                >>> coord_space = ['x', 'y', 'z']
                >>> coord_space = {"names": ['x', 'y', 'z']}
                >>> coord_space = {
                    "names": ['x', 'y', 'z'],
                    "units": ['nm', 'nm', 'nm'],
                    "scales": [1, 1, 1]
                }
                >>> coord_space = CoordinateSpace(
                ...     names=['x', 'y', 'z'],
                ...     scales=[1.0, 1.0, 1.0],
                ...     units=['nm', 'nm', 'nm']
                ... )

        annotation_type:
            Literal['point', 'line', 'ellipsoid', 'axis_aligned_bounding_box']
            The type of annotation to export. Note that the columns you provide in
            the DataFrame depend on the annotation type.

        properties:
            If your dataframe contains columns for annotation properties,
            list the names of those columns here.

            Categorical columns will be automatically converted to integers with associated
            enum labels.

            To provide an rgb or rgba property such as 'mycolor', provide separate columns
            in your dataframe named 'mycolor_r', 'mycolor_g', 'mycolor_b' (and 'mycolor_a'),
            and then include 'mycolor' in the properties list here.

            The full property spec for each property will be inferred from the column dtype,
            but if you want to explicitly override any property specs yourself, you can pass
            a list of AnnotationPropertySpec objects here instead of just listing column names.

            Property names must start with a lowercase letter and may contain only letters,
            numbers, and underscores.

        relationships:
            list[str]
            If your annotations have related segment IDs, such relationships can be provided
            in the columns of your DataFrame. Each relationship should be listed in a single column,
            whose values are lists of segment IDs.  In the special case where each annotation has
            exactly one related segment, the column may have dtype=np.uint64 instead of containing lists.

        output_dir:
            str
            The directory into which the exported annotations will be written.
            Subdirectories will be created for the "annotation ID index" and each
            "related object id index" as needed.

        write_sharded:
            bool
            Whether to write the output as sharded files.
            The sharded format is preferable for most use cases.
            Without sharding, every annotation results in a separate file in the annotation ID index.
            Similarly, every related ID results in a separate file in the related ID index.

        write_by_id:
            bool
            Whether to write the annotations to the "Annotation ID Index".
            If False, skip writing.

        write_relationships:
            bool
            Whether to write the relationships to the "Related Object ID Index".
            If False, skip writing.

        write_by_spatial_chunk:
            bool
            Whether to write the spatial index.

        num_spatial_levels:
            int
            The maximum number of spatial index levels to write.
            If not all levels are needed (because all annotations fit within the first N levels),
            then the actual number of levels written will be less than this value.

        target_chunk_limit:
            int
            For the spatial index, this is how many annotations we aim to place in each
            chunk (regardless of the level).
            If there are more annotations than fit within the specified num_spacial_levels
            while (approximately) adhering to the target_chunk_limit at each level, then the
            extra annotations will be assigned to the last level.

            Note:
                Instead of specifying a valid limit here, you can disable subsampling in neuroglancer
                by setting this to the special value of 0.  In our implementation, this is only valid
                when num_spatial_levels=1.

        shuffle_before_assigning_spatial_levels:
            bool
            Whether to shuffle the annotations before assigning spatial levels.
            By default, we shuffle the annotations to avoid any bias in the spatial
            assignment, which is what the neuroglancer spec recommends.
            However, in some use-cases a bias may be desirable (e.g. deliberately
            preferring to show larger annotations when zoomed out).
            So if this is False, the annotations will be assigned to spatial levels in
            the order they appear in the input dataframe, with earlier annotations
            assigned to coarser spatial levels.
        
        description:
            str
            A description of the annotation collection.
    """
    if isinstance(df, TableHandle):
        # Take ownership of the dataframe.
        handle, df = df, df.df
        handle.df = None

    coord_space = _construct_coord_space(coord_space)
    annotation_type = annotation_type.lower()
    property_specs = annotation_property_specs(df, properties)
    bounds = _get_bounds(df, coord_space, annotation_type)
    output_dir = os.path.abspath(output_dir)
    os.makedirs(output_dir, exist_ok=True)

    # Construct a buffer for each annotation and additional buffers
    # for each annotation's relationships, stored in new columns of df.
    df = _encode_annotations(
        df,
        coord_space,
        annotation_type,
        property_specs,
        relationships
    )

    by_id_metadata = {}
    if write_by_id:
        by_id_metadata = _write_annotations_by_id(
            df,
            output_dir,
            write_sharded
        )

    if write_by_relationship:
        df_handle_for_rel = TableHandle(df)

    if write_by_spatial_chunk:
        df_handle_for_spatial = TableHandle(df)

    # Delete our reference to df.
    # The TableHandles own the data now.
    del df

    by_rel_metadata = []
    if write_by_relationship:    
        by_rel_metadata = _write_annotations_by_relationships(
            df_handle_for_rel,
            relationships,
            output_dir,
            write_sharded
        )

    spatial_metadata = []
    if write_by_spatial_chunk:
        spatial_metadata = _write_annotations_by_spatial_chunk(
            df_handle_for_spatial,
            coord_space,
            annotation_type,
            bounds,
            num_spatial_levels,
            target_chunk_limit,
            shuffle_before_assigning_spatial_levels,
            output_dir,
            write_sharded
        )
    
    # Write the top-level 'info' file for the annotation output directory.
    # Check for NaN in bounds (which would produce invalid JSON)
    if np.any(np.isnan(bounds[0])) or np.any(np.isnan(bounds[1])):
        raise ValueError(
            f"Bounds contain NaN values: lower={bounds[0]}, upper={bounds[1]}. "
            "Check your input data for missing or invalid coordinate values."
        )

    info = {
        "@type": "neuroglancer_annotations_v1",
        "dimensions": coord_space.to_json(),
        "lower_bound": bounds[0].tolist(),
        "upper_bound": bounds[1].tolist(),
        "annotation_type": annotation_type,
        "properties": property_specs,
        "by_id": by_id_metadata,
        "relationships": by_rel_metadata,
        "spatial": spatial_metadata,
    }

    if description:
        info['description'] = description

    with open(f"{output_dir}/info", 'w') as f:
        json.dump(info, f)


def _construct_coord_space(coord_space):
    """
    This function produces a CoordinateSpace object from any of our accepted
    formats as explained in the docs for write_precomputed_annotations().

    Returns:
        CoordinateSpace
    """
    if isinstance(coord_space, CoordinateSpace):
        return coord_space

    if isinstance(coord_space, str):
        if coord_space != coord_space.lower() or len(set(coord_space)) != len(coord_space):
            raise ValueError(f"Invalid coordinate space: {coord_space!r}.")
        return CoordinateSpace(
            names=list(coord_space),
            units=['nm']*len(coord_space),
            scales=[1]*len(coord_space),
        )

    if isinstance(coord_space, list):
        if not all(isinstance(c, str) and c == c.lower() for c in coord_space):
            raise ValueError(f"Invalid coordinate space: {coord_space!r}.")
        return CoordinateSpace(
            names=coord_space,
            units=['nm']*len(coord_space),
            scales=[1]*len(coord_space),
        )

    if isinstance(coord_space, dict):
        if 'names' not in coord_space:
            return CoordinateSpace(json=coord_space)

        if not (coord_space.keys() <= {'names', 'units', 'scales', 'coordinate_arrays'}):
            raise ValueError(f"Invalid coordinate space: {coord_space!r}.")

        default_coord_space = {
            'names': coord_space['names'],
            'units': ['nm']*len(coord_space['names']),
            'scales': [1]*len(coord_space['names']),
        }
        return CoordinateSpace(**(default_coord_space | coord_space))

    raise ValueError(f"Invalid coordinate space: {coord_space!r}.")


def _get_bounds(df, coord_space, annotation_type):
    """
    Inspect the geometry columns of the given dataframe to
    determine the overall upper and lower bounds of the annotations.

    Returns:
        lower_bound, upper_bound
        (both numpy arrays of length 3)
    """
    geometry_cols = _geometry_cols(coord_space.names, annotation_type)
    if not (required_cols := set(chain(*geometry_cols))) <= set(df.columns):
        raise ValueError(
            "Dataframe does not have all required geometry columns for the given coordinate space.\n"
            f"Required columns: {required_cols}"
        )

    if annotation_type == 'point':
        points = df[geometry_cols[0]]
        return (
            points.min(axis=0).to_numpy(),
            points.max(axis=0).to_numpy()
        )

    if annotation_type in ('line', 'axis_aligned_bounding_box'):
        points_a = df[geometry_cols[0]]
        points_b = df[geometry_cols[1]]
        return (
            np.minimum(points_a.min().to_numpy(), points_b.min().to_numpy()),
            np.maximum(points_a.max().to_numpy(), points_b.max().to_numpy())
        )

    if annotation_type == 'ellipsoid':
        center = df[geometry_cols[0]].to_numpy()
        radii = df[geometry_cols[1]].to_numpy()
        return np.asarray([
            (center - radii).min(axis=0),
            (center + radii).max(axis=0)
        ])

    raise ValueError(f"Annotation type {annotation_type} not supported")


def _encode_annotations(df, coord_space, annotation_type, property_specs, relationships):
    """
    Returns a (shallow) copy of the dataframe with additional columns containing
    buffers for each encoded annotation and its encoded id and encoded relationships.
    """
    df = df.copy(deep=False)

    logger.info("Encoding annotation IDs")
    df['id_buf'] = _encode_uint64_series(df.index)

    logger.info("Encoding annotation geometries and properties")
    df['ann_buf'] = _encode_geometries_and_properties(
        df, coord_space, annotation_type, property_specs
    )

    logger.info("Encoding relationships")
    rel_bufs = _encode_relationships(df, relationships)
    if rel_bufs is not None:
        df['rel_buf'] = rel_bufs

    return df


def _encode_geometries_and_properties(df, coord_space, annotation_type, property_specs):
    """
    For each annotation in the given dataframe, encode its geometry columns (e.g. x,y,z)
    and property columns into a buffer, plus any padding that was necessary to align the
    buffer to a 4-byte boundary, per the neuroglancer spec.

    (In the precomputed format, geometry and properties always appear together,
    regardless of whether they're being written to the "Annotation ID Index",
    the "Related Object ID Index" or the "Spatial Index".)

    Returns:
        pd.Series of dtype=object, containing one buffer for each annotation.
    """
    geometry_cols = _geometry_cols(coord_space.names, annotation_type)
    geometry_prop_df = _geometry_prop_df(df, geometry_cols, property_specs)
    buf, recsize = _encode_geometry_prop_df(geometry_prop_df, geometry_cols, property_specs)
    del geometry_prop_df

    # extract bytes from the appropriate slice for each record
    encoded_annotations = [buf[i*recsize:(i+1)*recsize] for i in range(len(df))]
    ann_bufs = pd.Series(encoded_annotations, index=df.index)
    return ann_bufs


def _geometry_prop_df(df, geometry_cols, property_specs):
    """
    Select the subset of columns that specify the geometry and properties
    of the annotations, and append columns for padding that will ensure
    our encoded records have a width that is a multiple of 4 bytes.
    """
    # Note that the property specs are already sorted by
    # dtype in the order neuroglancer requires.
    # Order our columns accordingly before we encode them as records below.
    prop_cols = []
    for spec in property_specs:
        p = spec['id']
        if spec['type'] == 'rgb':
            prop_cols.extend([f'{p}_r', f'{p}_g', f'{p}_b'])
        elif spec['type'] == 'rgba':
            prop_cols.extend([f'{p}_r', f'{p}_g', f'{p}_b', f'{p}_a'])
        else:
            prop_cols.append(p)

    geometry_prop_df = df[[*chain(*geometry_cols), *prop_cols]].copy(deep=False)

    # Calculate padding as required by neuroglancer.
    property_widths = {}
    for spec in property_specs:
        if spec['type'] == 'rgb':
            property_widths[spec['id']] = 3
        elif spec['type'] == 'rgba':
            property_widths[spec['id']] = 4
        else:
            property_widths[spec['id']] = np.dtype(spec['type']).itemsize

    property_padding = (4 - (sum(property_widths.values()) % 4)) % 4
    for i in range(property_padding):
        geometry_prop_df[f'__padding_{i}__'] = np.uint8(0)

    return geometry_prop_df


def _encode_geometry_prop_df(geometry_prop_df, geometry_cols, property_specs):
    """
    Encode the geometry and properties of the annotations into a single buffer.

    Note: Replaces category columns of geometry_prop_df with their integer equivalents.
    """
    # Convert our column dtypes to match property specs.
    dtypes = {c: np.float32 for c in chain(*geometry_cols)}
    for spec in property_specs:
        p = spec['id']
        if spec['type'] in ('rgb', 'rgba'):
            dtypes.update({
                f'{p}_{channel}': np.uint8
                for channel in spec['type']
            })
        else:
            dtypes[p] = spec['type']

    if any(dt == np.int8 for dt in dtypes.values()):
        logger.warning(
            "Old versions of neuroglancer don't support int8 properties, "
            "so consider casting to uint8 or int16 if your annotations don't load."
        )

    # Convert category columns to their integer equivalents
    for spec in property_specs:
        p = spec['id']
        if spec['type'] in ('rgb', 'rgba'):
            continue
        if geometry_prop_df[p].dtype == 'category':
            geometry_prop_df[p] = geometry_prop_df[p].cat.codes
            dtypes[p] = spec['type']

    # Vectorized serialization
    records = geometry_prop_df.to_records(index=False, column_dtypes=dtypes)
    recsize = records.dtype.itemsize
    buf = records.tobytes()
    return buf, recsize



"""
Python implementation of neuroglancer's pseudo-random segment color generator.

Adapted from `Austin Hoag's implementation <https://github.com/google/neuroglancer/blob/master/python/neuroglancer/segment_colors.py>`_.
This version is lightly refactored to support JIT-compilation and vectorization using numba.

See also:
The source code here in segmentcolors.py also includes a WebGL implementation
of the same algorithm, suitable for neuroglancer annotation shaders.
"""
# @license
# Copyright 2017 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from numba import njit, vectorize

# Random note: Some ID pairs such as [818314, 800181] seem to have very close colors regardless of seed.


def hex_string_from_segment_id(color_seed, segment_id):
    """
    Return the hex color string for a segment
    given a color seed and the segment id.

    If color_seed and/or segment_id is an array, then they'll be
    broadcasted together and a list of colors will be returned.
    """
    packed_color = packed_color_from_segment_id(color_seed, segment_id)
    if hasattr(packed_color, '__len__'):
        return [f"#{c:06x}" for c in packed_color]
    return f"#{packed_color:06x}"


def rgb_from_segment_id(color_seed, segment_id):
    """
    Hash the segment_id to an RGB array of uint8 according to the given color_seed.

    If color_seed and/or segment_id is an array, then they'll be
    broadcasted together and a list of colors will be returned.
    """
    packed_color = packed_color_from_segment_id(color_seed, segment_id)
    return unpack_color(packed_color)


def unpack_color(packed_color):
    """
    Unpack an array of packed colors with shape (...,) into an
    array of shape (..., 3) with r,g,b in the last dimension.
    Also works for a single packed color scalar.
    """
    r = packed_color >> 16
    g = (packed_color >> 8) & 0xFF
    b = packed_color & 0xFF
    return np.concatenate(
        [
            r[..., None],
            g[..., None],
            b[..., None],
        ],
        axis=-1
    )


@njit
def _rgb_from_segment_id(color_seed, segment_id):
    segment_id = int(segment_id)  # necessary since segment_id is 64 bit originally
    result = hash_function(state=color_seed,value=segment_id)
    newvalue = segment_id >> 32
    result2 = hash_function(state=result,value=newvalue)
    c0 = (result2 & 0xFF) / 255.
    c1 = ((result2 >> 8) & 0xFF) / 255.
    h = c0
    s =  0.5 + 0.5 * c1
    v = 1.0
    rgb = hsv_to_rgb(h,s,v)
    return rgb


@njit
def pack_color(rgb_vec):
    """
    Returns an integer formed
    by concatenating the channels of the input color vector.
    Python implementation of packColor in src/neuroglancer/util/color.ts
    """
    result = 0
    for c in rgb_vec:
        result = ((result << 8) & 0xffffffff) + min(255,max(0,round(c*255)))
    return result


@njit
def hash_function(state,value):
    """ Python implementation of hashCombine() function
    in src/neuroglancer/gpu_hash/hash_function.ts,
    a modified murmur hash
    """
    k1 = 0xcc9e2d51
    k2 = 0x1b873593
    state = state & 0xffffffff
    value = (value * k1) & 0xffffffff
    value = ((value << 15) | value >> 17) & 0xffffffff
    value = (value * k2) & 0xffffffff
    state = (state ^ value) & 0xffffffff
    state = (( state << 13) | state >> 19) & 0xffffffff
    state = (( state * 5) + 0xe6546b64) & 0xffffffff
    return state


@njit
def hsv_to_rgb(h,s,v):
    """
    Convert H,S,V values to RGB values.
    Python implementation of hsvToRgb in src/neuroglancer/util/colorspace.ts
    """
    h *= 6
    hue_index = np.floor(h)
    remainder = h - hue_index
    val1 = v*(1-s)
    val2 = v*(1-(s*remainder))
    val3 = v*(1-(s*(1-remainder)))
    hue_remainder = hue_index % 6
    if hue_remainder == 0:
        return (v,val3,val1)
    elif hue_remainder == 1:
        return (val2,v,val1)
    elif hue_remainder == 2:
        return (val1,v,val3)
    elif hue_remainder == 3:
        return (val1,val2,v)
    elif hue_remainder == 4:
        return (val3,val1,v)
    elif hue_remainder == 5:
        return (v,val1,val2)


@vectorize("int64(int64, int64)", nopython=True)
def packed_color_from_segment_id(color_seed, segment_id):
    rgb = _rgb_from_segment_id(color_seed, segment_id)
    return pack_color(rgb_vec=rgb)


# For annotation shaders, here's a WebGL implementation (which only supports 32-bit segment IDs):
_SHADER_IMPLEMENTATION_ = """\
#uicontrol int color_seed slider(min=0, max=4294967295, default=0);
#uicontrol int segment_id slider(min=0, max=4294967295, default=1);

uint hash_function(uint state, uint value) {
    uint k1 = 0xcc9e2d51u;
    uint k2 = 0x1b873593u;

    value = value * k1;
    value = (value << 15) | (value >> 17);
    value = value * k2;
    state = state ^ value;
    state = (state << 13) | (state >> 19);
    state = state * 5u + 0xe6546b64u;

    return state;
}

vec3 hsv_to_rgb(float h, float s, float v) {
    h *= 6.0;
    float hue_index = floor(h);
    float remainder = h - hue_index;
    float val1 = v * (1.0 - s);
    float val2 = v * (1.0 - (s * remainder));
    float val3 = v * (1.0 - (s * (1.0 - remainder)));

    int hue_remainder = int(hue_index) % 6;

    if (hue_remainder == 0) {
        return vec3(v, val3, val1);
    } else if (hue_remainder == 1) {
        return vec3(val2, v, val1);
    } else if (hue_remainder == 2) {
        return vec3(val1, v, val3);
    } else if (hue_remainder == 3) {
        return vec3(val1, val2, v);
    } else if (hue_remainder == 4) {
        return vec3(val3, val1, v);
    } else { // hue_remainder == 5
        return vec3(v, val1, val2);
    }
}

vec3 color_from_segment_id(uint color_seed, uint segment_id) {
    uint result = hash_function(color_seed, segment_id);

    // We only support 32-bit segment IDs, so newvalue is always 0,
    // and (id >> 32) would be undefined behavior!
    // uint newvalue = id >> 32;

    uint newvalue = 0u;
    uint result2 = hash_function(result, newvalue);

    float c0 = float(result2 & 0xFFu) / 255.0;
    float c1 = float((result2 >> 8) & 0xFFu) / 255.0;

    float h = c0;
    float s = 0.5 + 0.5 * c1;
    float v = 1.0;

    return hsv_to_rgb(h, s, v);
}

void main() {
  setPointMarkerSize(10.0);
  setPointMarkerBorderColor(vec4(0, 0, 0, 1.0));

  vec3 hash_color = color_from_segment_id(uint(color_seed), uint(segment_id));
  setColor(vec4(hash_color, 1.0));
}
"""



#!/usr/bin/env python
"""
Simple CORS-enabled web server for serving local files to Neuroglancer.

This server supports HTTP Range requests, which are required for
neuroglancer's sharded precomputed format.

WARNING: Because this web server permits cross-origin requests, it exposes any
data in the directory that is served to any web page running on a machine that
can connect to the web server.
"""

import argparse
import os
import sys
import mimetypes
from flask import Flask, send_from_directory, after_this_request, request
from werkzeug.exceptions import NotFound


def create_app(directory):
    """Create a Flask app that serves files from the given directory."""
    app = Flask(__name__)
    
    # Add custom MIME types for neuroglancer
    mimetypes.add_type('application/json', '')  # Files without extension
    mimetypes.add_type('application/octet-stream', '.shard')
    
    @app.route('/', defaults={'path': ''})
    @app.route('/<path:path>')
    def serve_file(path):
        """Serve files with CORS headers. Flask automatically handles Range requests."""
        # Log the request
        range_header = request.headers.get('Range', 'no range')
        print(f"Request: {path} [{range_header}]", file=sys.stderr)
        
        @after_this_request
        def add_cors_headers(response):
            response.headers['Access-Control-Allow-Origin'] = '*'
            response.headers['Access-Control-Allow-Headers'] = 'Range'
            response.headers['Access-Control-Expose-Headers'] = 'Content-Length, Content-Range, Accept-Ranges'
            print(f"Response: {response.status_code} for {path}", file=sys.stderr)
            return response
        
        try:
            # send_from_directory automatically handles Range requests
            return send_from_directory(directory, path if path else '.')
        except NotFound:
            # Try serving as index
            if not path or path.endswith('/'):
                return send_from_directory(directory, 'index.html')
            raise
    
    return app


if __name__ == "__main__":
    ap = argparse.ArgumentParser(
        description="Serve local files with CORS and Range request support for Neuroglancer"
    )
    ap.add_argument(
        "-p", "--port", type=int, default=9000, help="TCP port to listen on"
    )
    ap.add_argument("-a", "--bind", default="127.0.0.1", help="Bind address")
    ap.add_argument("-d", "--directory", default=".", help="Directory to serve")

    args = ap.parse_args()
    
    directory = os.path.abspath(args.directory)
    print(f"Serving directory {directory} at http://{args.bind}:{args.port}")
    
    app = create_app(directory)
    
    try:
        app.run(host=args.bind, port=args.port, threaded=True)
    except KeyboardInterrupt:
        print("\nShutting down server...")
        sys.exit(0)



import logging
import numpy as np
import pandas as pd

from ._util import _encode_uint64_series, TableHandle
from ._write_buffers import _write_buffers

logger = logging.getLogger(__name__)


def _encode_relationships(df, relationships):
    """
    For each annotation in the given dataframe, encode the related IDs
    for all relationships into a buffer according to the neuroglancer spec.

    Returns:
        pd.Series of dtype=object, containing one buffer for each annotation.
    """
    if not relationships:
        return None

    encoded_relationships = {}
    for rel_col in relationships:
        encoded_relationships[rel_col] = _encode_related_ids(df[rel_col])

    # Concatenate buffers on each row.
    # Note:
    #   Using sum() is O(R^2) in the number of relationships R, but we generally
    #   expect few relationships, so this is faster than df.apply(b''.join, axis=1),
    #   since .sum() uses a single C call whereas using b''.join() would
    #   use many Python calls.
    rel_bufs = pd.DataFrame(encoded_relationships, index=df.index).sum(axis=1)
    return rel_bufs


def _encode_related_ids(related_ids):
    """
    Given a Series containing lists of IDs, encode each list of IDs
    in the format neuroglancer expects for each relationship in an
    annotation.

    Each item in related_ids is a list, which gets encoded as
    <count><id_1><id_2><id_3>..., where <count> is uint32 and
    <id_1><id_2><id_3>... are each uint64.

    Args:
        related_ids:
            A Series of length N and dtype=object, containing lists of IDs.
            As a special convenience in the case where every row contains
            exactly one ID, you may pass a series with dtype=auint64,
            which will be interpreted as if each entry were a list of length 1.
            (In this case, the implementation is slightly faster than in the general case.)

    Returns:
        A numpy array with N entries, where each entry is a buffer as shown above.
    """
    # Special case if the relationship contains only a single ID for each annotation.
    if np.issubdtype(related_ids.dtype, np.integer):
        buf = (
            pd.DataFrame({'count': np.uint32(1), 'id': related_ids})
            .astype({'count': np.uint32, 'id': np.uint64}, copy=False)
            .to_records(index=False)
            .tobytes()
        )

        encoded_ids = [
            buf[i*12:(i+1)*12]
            for i in range(len(related_ids))
        ]
        return np.array(encoded_ids, dtype=object)

    # Otherwise, the relationship contains lists.
    else:
        assert related_ids.dtype == object
        counts = related_ids.map(len).to_numpy(np.uint32)
        offsets = 8 * np.cumulative_sum(counts, include_initial=True)

        ids_buf = np.concatenate(related_ids, dtype=np.uint64).tobytes()
        counts_buf = counts.tobytes()

        encoded_ids = [
            counts_buf[i*4:(i+1)*4] + ids_buf[start:end]
            for i, (start, end) in enumerate(zip(offsets[:-1], offsets[1:]))
        ]
        return np.array(encoded_ids, dtype=object)


def _write_annotations_by_relationships(df_handle: TableHandle, relationships, output_dir, write_sharded):
    """
    Write the annotations to a "Related Object ID Index" for each relationship.
    Each relationship is written to a separate subdirectory of output_dir.

    Args:
        df_handle:
            TableHandle holding a DataFrame with columns ['id_buf', 'ann_buf', *relationships].
            The handle's reference will be unset before this function returns.

        relationships:
            List of relationship column names.

        output_dir:
            Directory to write the annotations to.
            Each relationship is written to a separate subdirectory of output_dir.

        write_sharded:
            Whether to write the annotations in sharded format.

    Returns:
        JSON metadata to be written under the 'relationships' key in the top-level 'info' file,
        consisting of a list of JSON objects (one for each relationship).
    """
    handles = {
        r: TableHandle(df_handle.df) for r in relationships
    }
    df_handle.df = None

    by_rel_metadata = []
    for relationship, df_handle in handles.items():
        metadata = _write_annotations_by_relationship(
            df_handle,
            relationship,
            output_dir,
            write_sharded
        )
        by_rel_metadata.append(metadata)

    return by_rel_metadata


def _write_annotations_by_relationship(df_handle: TableHandle, relationship, output_dir, write_sharded):
    """
    Write the annotations to a "Related Object ID Index" for a single relationship.

    Returns:
        JSON metadata for the relationship, including the key and sharding spec if applicable.
    """
    logger.info(f"Grouping annotations by relationship {relationship}")
    bufs_by_segment = (
        df_handle.df[['id_buf', 'ann_buf', relationship]]
        .dropna(subset=relationship)
        .explode(relationship)
        .groupby(relationship, sort=False)
        # Use b''.join() instead of 'sum' to avoid O(N^2) performance for large groups.
        .agg({'id_buf': ['count', b''.join], 'ann_buf': b''.join})
    )
    df_handle.df = None

    logger.info(f"Combining annotation and ID buffers for relationship '{relationship}'")
    bufs_by_segment.columns = ['count', 'id_buf', 'ann_buf']
    bufs_by_segment['count_buf'] = _encode_uint64_series(bufs_by_segment['count'])
    bufs_by_segment['combined_buf'] = bufs_by_segment[['count_buf', 'ann_buf', 'id_buf']].sum(axis=1)

    logger.info(f"Writing annotations to 'by_rel_{relationship}' index")
    metadata = _write_buffers(
        bufs_by_segment['combined_buf'],
        output_dir,
        f"by_rel_{relationship}",
        write_sharded
    )
    metadata['id'] = relationship
    return metadata



import logging
from itertools import chain
from typing import NamedTuple

import numpy as np
from numba import njit
from numba.typed import List

from .compressed_morton import compressed_morton_code, compressed_morton_decode, compressed_morton_code_no_broadcast
from ._write_buffers import _write_buffers
from ._util import _encode_uint64_series, _geometry_cols, _ndindex_array, TableHandle

logger = logging.getLogger(__name__)

GridSpec = NamedTuple("GridSpec", [('chunk_shapes', np.ndarray), ('grid_shapes', np.ndarray)])


def _write_annotations_by_spatial_chunk(
        df_handle: TableHandle,
        coord_space,
        annotation_type,
        bounds,
        num_levels,
        target_chunk_limit,
        shuffle_before_assigning_spatial_levels,
        output_dir,
        write_sharded
):
    """
    Write the annotations to the spatial index.

    Args:
        df_handle:
            TableHandle.  The handle's reference will be unset before this function returns.
            The enclosed DataFrame must have columns ['id_buf', 'ann_buf', *geometry_cols].
            Internally, the data will be copied during processing and again
            during writing, incurring significant RAM usage for large datasets.
            The caller can save some RAM by deleting their own reference to the input
            after constructing the TableHandle (but before calling this function).

        coord_space:
            CoordinateSpace.
            The coordinate space of the annotations.

        annotation_type:
            Literal['point', 'axis_aligned_bounding_box', 'ellipsoid', 'line']
            The type of annotation to export.

        bounds:
            np.ndarray, shape (2, D)
            Lower and upper bounds of the union of all annotations.
            The bounds are in coordinate units.

        num_levels:
            The number of spatial index levels. Must be at least 1.

        target_chunk_limit:
            The maximum number of annotations to place in each chunk.
            (The same target is used for all levels.)
            Since the spatial annotations are not distributed uniformly in space,
            we will likely end up undershooting and overshooting the target for various
            chunks within a level.
            The final maximum number of annotations per chunk we end up with at each
            level will be emitted in the the 'limit' setting of the metadata for each level.

            Note:
                Instead of specifying a valid limit here, you can disable subsampling in neuroglancer
                by setting this to the special value of 0.  This is only valid when num_levels=1.

        shuffle_before_assigning_spatial_levels:
            Whether to shuffle the annotations before assigning spatial levels.
            If False, the annotations will be assigned to spatial levels in the order
            they appear in the input dataframe, with earlier annotations assigned to
            coarser spatial levels.
            By default, we shuffle the annotations to avoid any bias in the spatial
            assignment, which is what the neuroglancer spec recommends.

        output_dir:
            Directory to write the annotations to.
            Subdirectories for each level of the spatial index will be created in output_dir,
            named 'by_spatial_level_<level>'.

        write_sharded:
            Whether to write the annotations in sharded format.

    Returns:
        JSON metadata to write into the 'spatial' key of the info file.
    """
    df_handle, gridspec = _assign_spatial_chunks(
        df_handle,
        coord_space,
        annotation_type,
        bounds,
        num_levels,
        target_chunk_limit,
        shuffle_before_assigning_spatial_levels
    )

    metadata = _write_assigned_annotations_by_spatial_chunk(
        df_handle,
        gridspec,
        (target_chunk_limit == 0),
        output_dir,
        write_sharded
    )
    return metadata


def _assign_spatial_chunks(
    df_handle: TableHandle,
    coord_space,
    annotation_type,
    bounds,
    num_levels,
    target_chunk_limit,
    shuffle_before_assigning_spatial_levels
):
    """
    Assign each annotation to a spatial grid cell.
    If an annotation intersects multiple grid cells, we duplicate
    its row so we can assign it to all of the intersecting cells.

    Args:
        df_handle:
            TableHandle.  The handle's reference will be unset before this function returns.
        ...

    Returns:
        df, gridspec

        - df is a shuffled copy of the input df with all columns removed except
          'ann_buf' and 'id_buf', and with additional columns for 'level' and 'chunk_code'.
          Some rows from the original dataframe may be duplicated if those annotations
          span across multiple chunks (at the level we selected them to reside in).
        - gridspec: chunk_shapes and grid_shapes.  See _define_spatial_grids() for details.
    """
    df = df_handle.df
    df_handle.df = None

    geometry_cols = _geometry_cols(coord_space.names, annotation_type)
    df = df[[*chain(*geometry_cols), 'ann_buf', 'id_buf']].copy(deep=False)
    gridspec = _define_spatial_grids(bounds, coord_space, num_levels)
    level_annotation_counts = _compute_target_annotations_per_level(len(df), gridspec, target_chunk_limit)

    if shuffle_before_assigning_spatial_levels:
        logger.info("Shuffling annotations before assigning spatial grid levels")
        df = df.sample(frac=1.0)

    logger.info("Assigning spatial grid chunks")
    df['level'] = np.repeat(
        range(num_levels),
        level_annotation_counts.astype(int)
    ).astype(np.uint64)
    
    match annotation_type:
        case 'point':
            df = _assign_spatial_chunks_for_points(df, geometry_cols, bounds, gridspec)
        case 'axis_aligned_bounding_box':
            df = _assign_spatial_chunks_for_axis_aligned_bounding_boxes(df, geometry_cols, bounds, gridspec)
        case 'ellipsoid':
            df = _assign_spatial_chunks_for_ellipsoids(df, geometry_cols, bounds, gridspec)
        case 'line':
            df = _assign_spatial_chunks_for_lines(df, geometry_cols, bounds, gridspec)
        case _:
            raise NotImplementedError(f"Spatial indexing for {annotation_type} annotations is not implemented")

    logger.info("Done assigning spatial grid chunks")
    return TableHandle(df), gridspec


def _define_spatial_grids(bounds, coord_space, num_levels: int) -> GridSpec:
    """
    Compute suitable chunk shapes and grid shapes for each level 
    of the spatial index, following the guidelines from the spec[1]:

        > Typically the grid_shape for level 0 should be a vector of all 1
        > (with chunk_size equal to upper_bound - lower_bound), and each component
        > of chunk_size of each successively level should be either equal to, or half of,
        > the corresponding component of the prior level chunk_size, whichever results
        > in a more spatially isotropic chunk.

    [1]: https://github.com/google/neuroglancer/blob/master/src/datasource/precomputed/annotations.md#spatial-index

    Args:
        bounds:
            np.ndarray, shape (2, D)
            lower and upper bounds of the union of all annotations

        coord_space:
            Needed to aim for roughly isotropic chunks in physical units.

        num_levels:
            The number of spatial index levels. Must be at least 1.

    Returns:
        GridSpec(chunk_shapes, grid_shapes)

        - chunk_shapes is the array (for N levels) of the size of each
          grid cell at the corresponding level, in coordinate units.
        - grid_shapes is the array (for N levels) of the number of grid cells
          along each dimension at the corresponding level.

        For instance, level 0 consists of a single chunk encompassing the entire
        volume occupied by the annotations, so its chunk_shape is the entire bounds
        (offset by the lower bound) and its grid_shape is [1,1,...].
    """
    # Level 0 chunk shape and grid shape -- just one chunk.
    bounds = np.asarray(bounds, np.float64)

    # We want roughly isotropic chunks in physical units, so we'll multiply
    # by the coordinate scales and then divide the scales out at the end.
    chunk_shape = (bounds[1] - bounds[0]) * coord_space.scales
    grid_shape = np.ones_like(chunk_shape, dtype=np.uint64)

    chunk_shapes = [chunk_shape]
    grid_shapes = [grid_shape]

    for level in range(1, num_levels):
        chunk_shape = chunk_shape.copy()
        grid_shape = grid_shape.copy()

        max_dim = np.argmax(chunk_shape)
        target_width = chunk_shape[max_dim] / 2

        for dim, dim_width in enumerate(chunk_shape):
            if dim == max_dim:
                # Always split across the widest dimension.
                chunk_shape[dim] = target_width
                grid_shape[dim] *= 2
            elif (dim_width / target_width) > 1.5:
                # Split across this dimension to make it more isotropic.
                chunk_shape[dim] = dim_width / 2
                grid_shape[dim] *= 2
            else:
                # Splitting would make it less isotropic,
                # so leave this dimension unsplit.
                chunk_shape[dim] = dim_width
                grid_shape[dim] *= 1

        chunk_shapes.append(chunk_shape)
        grid_shapes.append(grid_shape)

    # Convert from physical units back to coordinate units.
    chunk_shapes = np.array(chunk_shapes, dtype=np.float64) / coord_space.scales
    grid_shapes = np.array(grid_shapes, dtype=np.uint64)

    return GridSpec(chunk_shapes, grid_shapes)


def _compute_target_annotations_per_level(num_annotations, gridspec, target_chunk_limit: int):
    """
    Compute the TOTAL number of annotations at each level of the spatial index.
    The target_chunk_limit is how many annotations we aim to place in each chunk
    (regardless of the level).
    
    Since the spatial annotations are not necessarily distributed uniformly in space,
    we will likely end up undershooting and overshooting the target for various
    chunks within a level.

    Furthermore, since the number of annotations passed in here is based on the
    table BEFORE duplicating annotations which span multiple chunks, the number
    of annotations at each level will eventually be more than what is returned here,
    after the appropriate duplications.

    Returns:
        np.ndarray, shape (num_levels,)
    """
    num_levels = len(gridspec.grid_shapes)
    chunk_counts_by_level = np.prod(gridspec.grid_shapes, axis=1)

    if target_chunk_limit != 0:
        annotation_counts = chunk_counts_by_level * target_chunk_limit
    else:
        assert num_levels == 1, \
            "The special target_chunk_limit of 0 is only permitted when num_spatial_levels=1"
        assert chunk_counts_by_level.tolist() == [1]
        annotation_counts = np.array([num_annotations])
    
    # Clamp to total number of annotations remaining after earlier levels
    for level in range(num_levels - 1):
        annotation_counts[level] = min(
            annotation_counts[level],
            num_annotations - sum(annotation_counts[:level])
        )

    # Last level gets all remaining annotations, if any.
    annotation_counts[-1] = num_annotations - sum(annotation_counts[:-1])

    return annotation_counts


def _assign_spatial_chunks_for_points(df, geometry_cols, bounds, gridspec):
    coord_names = geometry_cols[0]
    grid_indices = (df[[*coord_names]] - bounds[0]) // gridspec.chunk_shapes[df['level']]
    grid_indices = grid_indices.astype(np.int32)

    # Make sure annotations at the exact upper bound get valid grid coordinates.
    grid_indices = np.minimum(grid_indices, gridspec.grid_shapes[df['level']].astype(np.int32) - 1)

    # Switch to C order before computing compressed morton code.
    df['chunk_code'] = compressed_morton_code(
        grid_indices.to_numpy()[:, ::-1],
        gridspec.grid_shapes[df['level']][:, ::-1]
    )
    
    # FIXME: remove this debug code or put it somewhere else.
    # import pyarrow.feather as feather
    # df[[f'grid_{c}' for c in coord_names]] = grid_indices
    # df[[f'{c}a' for c in coord_names]] =  grid_indices * gridspec.chunk_shapes[df['level']] + bounds[0]
    # df[[f'{c}b' for c in coord_names]] = (grid_indices + 1) * gridspec.chunk_shapes[df['level']] + bounds[0]
    # feather.write_feather(df.drop(columns=['id_buf', 'ann_buf']), '/tmp/points.feather')
    
    return df[['level', 'chunk_code', 'id_buf', 'ann_buf']]


def _assign_spatial_chunks_for_axis_aligned_bounding_boxes(df, geometry_cols, bounds, gridspec):
    boxes = df[[*geometry_cols[0], *geometry_cols[1]]].to_numpy().reshape(len(df), 2, -1)

    # Ensure start < end
    swap_mask = (boxes[:, 0, :] > boxes[:, 1, :])[:, None, :]
    swap_mask = np.concatenate([swap_mask, swap_mask], axis=1)
    boxes[swap_mask] = boxes[:, ::-1, :][swap_mask]

    chunk_shapes = gridspec.chunk_shapes[df['level']]

    # FIXME: would be faster to compute the spans in _grid_box_codes()
    grid_spans = np.zeros_like(boxes, np.uint64)
    grid_spans[:, 0] = np.floor((boxes[:, 0] - bounds[0]) / chunk_shapes).astype(np.uint64)
    grid_spans[:, 1] = np.ceil((boxes[:, 1] - bounds[0]) / chunk_shapes).astype(np.uint64)

    df[f'chunk_code'] = _box_grid_codes(grid_spans, gridspec.grid_shapes[df['level']])

    # Duplicate the annotations which span multiple chunks.
    df = df[['level', 'chunk_code', 'id_buf', 'ann_buf']].explode('chunk_code')
    return df


@njit
def _box_grid_codes(grid_spans, grid_shapes):
    lists = List()
    for grid_span, grid_shape in zip(grid_spans, grid_shapes):
        grid_indices = grid_span[0] + _ndindex_array(grid_span[1] - grid_span[0])

        # Switch to C order before computing compressed morton code.
        codes = compressed_morton_code_no_broadcast(grid_indices[:, ::-1], grid_shape[::-1])
        lists.append(List(codes))
    return lists


def _assign_spatial_chunks_for_ellipsoids(df, geometry_cols, bounds, gridspec):
    centroids = df[geometry_cols[0]].to_numpy()
    radii = df[geometry_cols[1]].to_numpy()

    boxes = np.concatenate([centroids - radii, centroids + radii], axis=1).reshape(len(df), 2, -1)
    boxes = boxes - bounds[0]

    df[f'chunk_code'] = _ellipsoid_grid_codes(
        centroids,
        radii,
        df['level'].to_numpy(),
        bounds[0],
        gridspec.grid_shapes,
        gridspec.chunk_shapes
    )

    # Duplicate the annotations which span multiple chunks.
    df = df[['level', 'chunk_code', 'id_buf', 'ann_buf']].explode('chunk_code')
    return df


@njit
def _ellipsoid_grid_codes(centroids, radii, levels, grid_origin, grid_shapes, chunk_shapes):
    D = len(grid_shapes[0])
    lists = List()
    for centroid, radius, level in zip(centroids, radii, levels):
        grid_shape = grid_shapes[level]
        chunk_shape = chunk_shapes[level]

        grid_span = np.zeros((2, D), np.uint64)
        grid_span[0] = np.floor((centroid - radius - grid_origin) / chunk_shape)
        grid_span[1] = np.ceil((centroid + radius - grid_origin) / chunk_shape)

        grid_indices = grid_span[0] + _ndindex_array(grid_span[1] - grid_span[0])
        codes = List()
        for grid_index in grid_indices:
            if _ellipsoid_chunk_overlap(centroid, radius, grid_origin, chunk_shape, grid_index):
                # Switch to C order before computing compressed morton code.
                code = compressed_morton_code_no_broadcast(grid_index[::-1], grid_shape[::-1])
                codes.append(code)
        lists.append(codes)
    return lists


@njit
def _ellipsoid_chunk_overlap(center, radii, grid_origin, cell_shape, grid_index):
    """
    Ported from the C++ implementation[1] by jbms, except that we just return
    a boolean indicating whether the ellipsoid and cell have any overlap (True)
    or are completely disjoint (False).

    [1]: https://github.com/google/neuroglancer/pull/522#issuecomment-1940516294
    """
    rank = len(center)
    min_sum = 0.0

    for i in range(rank):
        cell_size = cell_shape[i]
        cell_start = grid_index[i] * cell_size + grid_origin[i]
        cell_end = cell_start + cell_size
        center_pos = center[i]
        
        start_dist = abs(cell_start - center_pos)
        end_dist = abs(cell_end - center_pos)
        
        if center_pos >= cell_start and center_pos <= cell_end:
            min_distance = 0.0
        else:
            min_distance = min(start_dist, end_dist)
        
        min_sum += min_distance**2 / radii[i]**2
    
    return min_sum <= 1.0


def _assign_spatial_chunks_for_lines(df, geometry_cols, bounds, gridspec):
    endpoints = df[[*geometry_cols[0], *geometry_cols[1]]].to_numpy().reshape(len(df), 2, -1)

    # Ensure start < end
    swap_mask = (endpoints[:, 0, :] > endpoints[:, 1, :])[:, None, :]
    swap_mask = np.concatenate([swap_mask, swap_mask], axis=1)
    endpoints[swap_mask] = endpoints[:, ::-1, :][swap_mask]

    df[f'chunk_code'] = _line_grid_codes(
        endpoints,
        df['level'].to_numpy(),
        bounds[0],
        gridspec.grid_shapes,
        gridspec.chunk_shapes
    )

    # Duplicate the annotations which span multiple chunks.
    df = df[['level', 'chunk_code', 'id_buf', 'ann_buf']].explode('chunk_code')
    return df


@njit
def _line_grid_codes(endpoints, levels, grid_origin, grid_shapes, chunk_shapes):
    D = len(grid_shapes[0])
    lists = List()
    for (point_a, point_b), level in zip(endpoints, levels):
        grid_shape = grid_shapes[level]
        chunk_shape = chunk_shapes[level]

        grid_span = np.zeros((2, D), np.uint64)
        grid_span[0] = np.floor((point_a - grid_origin) / chunk_shape)
        grid_span[1] = np.ceil((point_b - grid_origin) / chunk_shape)

        codes = List()
        grid_indices = grid_span[0] + _ndindex_array(grid_span[1] - grid_span[0])
        for grid_index in grid_indices:
            if _line_chunk_overlap(point_a, point_b, grid_origin, chunk_shape, grid_index):
                # Switch to C order before computing compressed morton code.
                code = compressed_morton_code_no_broadcast(grid_index[::-1], grid_shape[::-1])
                codes.append(code)
        lists.append(codes)
    return lists


@njit
def _line_chunk_overlap(point_a, point_b, grid_origin, cell_shape, grid_index):
    """
    Ported from the C++ implementation[1] by jbms.
    Returns True if the line intersects the cell, False otherwise.

    [1]: https://github.com/google/neuroglancer/pull/522#issuecomment-1940516294
    """
    rank = len(point_a)
    min_t = 0.0
    max_t = 1.0
    
    for i in range(rank):
        a = point_a[i]
        b = point_b[i]
        line_lower = min(a, b)
        line_upper = max(a, b)
        box_lower = grid_origin[i] + cell_shape[i] * grid_index[i]
        box_upper = box_lower + cell_shape[i]
        
        line_range = line_upper - line_lower
        
        if box_lower > line_lower:
            if line_range == 0.0:
                # Line is a point, check if it's outside the box
                if line_lower < box_lower:
                    return False
            else:
                t = (box_lower - line_lower) / line_range
                if t > 1:
                    return False
                min_t = max(min_t, t)
        
        if box_upper < line_upper:
            if line_range == 0.0:
                # Line is a point, check if it's outside the box
                if line_lower > box_upper:
                    return False
            else:
                t = (box_upper - line_lower) / line_range
                if t < 0:
                    return False
                max_t = min(max_t, t)
    
    return max_t >= min_t


def _write_assigned_annotations_by_spatial_chunk(df_handle, gridspec, disable_subsampling, output_dir, write_sharded):
    """
    Write the spatial index, given a dataframe in which the 'level'
    and grid chunk codes for each annotation have already been assigned.

    Args:
        df_handle:
            TableHandle.  The handle's reference will be unset before this function returns.
        gridspec:
            GridSpec object defining the spatial index.
        disable_subsampling:
            Whether to disable subsampling by seeting "limit" to 1 in the info file.
            (See inline comments.)
        output_dir:
            Directory to write the annotations to.
            Subdirectories will be created for each level of the spatial index.
        write_sharded:
            Whether to write the annotations in sharded format.

    Returns:
        JSON metadata to write into the 'spatial' key of the info file.
    """
    logger.info(f"Concatenating annotations by spatial chunk")
    bufs_by_grid = (
        df_handle.df[['level', 'chunk_code', 'id_buf', 'ann_buf']]
        .groupby(['level', 'chunk_code'], sort=False)
        .agg({'id_buf': ['count', b''.join], 'ann_buf': b''.join})
    )

    # We're done with the original input; delete it to save
    # RAM before writing (which takes a lot of RAM).
    df_handle.df = None

    logger.info(f"Combining annotation and ID buffers for spatial index")
    bufs_by_grid.columns = ['count', 'id_buf', 'ann_buf']
    bufs_by_grid['count_buf'] = _encode_uint64_series(bufs_by_grid['count'])
    bufs_by_grid['combined_buf'] = bufs_by_grid[['count_buf', 'ann_buf', 'id_buf']].sum(axis=1)

    bufs_by_grid = bufs_by_grid.reset_index()

    metadata = []
    for level, level_bufs in bufs_by_grid.groupby('level'):
        logger.info(f"Writing annotations to 'by_spatial_level_{level}' index")

        if write_sharded:
            # Sharded key is the compressed morton code of the grid coordinate.
            level_bufs.index = level_bufs['chunk_code']
        else:
            # Unsharded key is string of the grid coordinate, e.g. '0_0_0'
            grid_coords = compressed_morton_decode(
                level_bufs['chunk_code'].to_numpy(),
                gridspec.grid_shapes[level]
            )
            level_bufs.index = list(map('_'.join, grid_coords.astype(str)))

        level_metadata = _write_buffers(
            level_bufs['combined_buf'],
            output_dir,
            f"by_spatial_level_{level}",
            write_sharded
        )
        level_metadata['chunk_size'] = gridspec.chunk_shapes[level].tolist()
        level_metadata['grid_shape'] = gridspec.grid_shapes[level].tolist()

        if disable_subsampling:
            # To be honest, I don't completely understand why this
            # disables subsampling, but according to jbms[1]:
            #
            #   > Neuroglancer "subsamples" by showing only a prefix of the list of
            #   > annotations according to the spacing setting.  If you set "limit" to 1 in
            #   > the info file, you won't get subsampling by default.
            #
            # [1]: https://github.com/google/neuroglancer/issues/227#issuecomment-651944575
            level_metadata['limit'] = 1
        else:
            level_metadata['limit'] = int(level_bufs['count'].max())
        metadata.append(level_metadata)

    return metadata



"""
Utility functions related to storing neuroglancer states in Google Cloud Storage.
"""
import json
import argparse
import tempfile
import subprocess
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor


def upload_ngstates(bucket_dir, states, threads=0, processes=0, disable_cache=False, return_prefix='https://neuroglancer-demo.appspot.com'):
    """
    Use multithreading or multiprocessing to upload many files in parallel,
    similar to `gsutil -m cp []...]`, except that in this case you must choose
    between multithreading or multiprocessing (not a combination of the two).

    For the return values, neuroglancer links are returned (using the given prefix).
    If no prefix is given, then the URLs to the uploaded files are returned.

    Args:
        bucket_dir: str
            The bucket directory to upload the states to, such as 'gs://mybucket/mydir'
        states: dict
            A dictionary of neuroglancer states, keyed by the filename.
        threads: int
            The number of threads to use for uploading.
        processes: int
            The number of processes to use for uploading.
        disable_cache: bool
            If True, set `public, no-store` metadata on the uploaded files to disable caching.
        return_prefix: str
            If provided, return neuroglancer links using the given neuroglancer deployment.
            If None, then return the direct URLs to the uploaded files in google storage.

    Returns:
        list of str
            A list of links.
    """
    assert bucket_dir.startswith('gs://')
    assert processes == 0 or threads == 0, \
        "Pick multiprocessing or multithreading (or neither), but not both."

    bucket_dir = bucket_dir[len('gs://'):]
    bucket = bucket_dir.split('/')[0]
    dirpath = bucket_dir[1 + len(bucket):]

    blob_names = [dirpath + '/' + str(name) for name in states.keys()]
    blobs = map(json.dumps, states.values())
    args = [
        (bucket, blobname, blob, 'application/json', disable_cache)
        for blobname, blob in zip(blob_names, blobs)
    ]

    if processes > 0:
        with ProcessPoolExecutor(max_workers=processes) as executor:
            urls = list(executor.map(lambda args: upload_to_bucket(*args), args))
    elif threads > 0:
        with ThreadPoolExecutor(max_workers=threads) as executor:
            urls = list(executor.map(lambda args: upload_to_bucket(*args), args))
    else:
        # Sequential execution
        urls = [upload_to_bucket(*arg_tuple) for arg_tuple in args]

    if return_prefix:
        old_prefix = 'https://storage.googleapis.com/'
        assert all(url.startswith(old_prefix) for url in urls)
        return_prefix = return_prefix.rstrip('/') + '/#!gs://'
        urls = [url.replace(old_prefix, return_prefix) for url in urls]
    return urls


def upload_ngstate(bucket_path, state, disable_cache=False, return_prefix='https://neuroglancer-demo.appspot.com'):
    """
    Upload the given JSON state to a gbucket location,
    such as 'gs://flyem-user-links/short/foobar.json'

    For the return value, a neuroglancer link is returned (using the given prefix).
    If no prerix is given, then the URL to the uploaded file is returned.

    Args:
        bucket_path: str
            The bucket directory to upload the state to, such as 'gs://mybucket/mydir'
        state: dict
            The neuroglancer state to upload.
        disable_cache: bool
            If True, set `public, no-store` metadata on the uploaded file to disable caching.
        return_prefix: str
            If provided, return neuroglancer links using the given neuroglancer deployment.
            If None, then return the direct URLs to the uploaded files in google storage.

    Returns:
        str
            A neuroglancer link to the uploaded file or the direct URL to the
            uploaded file in google storage, depending on whether `return_prefix`
            was provided.
    """
    assert bucket_path.startswith('gs://')
    bucket_path = bucket_path[len('gs://'):]

    bucket = bucket_path.split('/')[0]
    filename = bucket_path[1 + len(bucket):]

    state_string = json.dumps(state, indent=2)
    url = upload_to_bucket(bucket, filename, state_string, disable_cache=disable_cache)

    if return_prefix:
        return_prefix = return_prefix.rstrip('/') + '/#!gs://'
        url = url.replace('https://storage.googleapis.com/', return_prefix)
    return url


def upload_json(obj, bucket_path, disable_cache=True):
    """
    Upload the given JSON data to a gbucket location,
    such as 'gs://flyem-user-links/short/foobar.json'
    """
    assert bucket_path.startswith('gs://')
    bucket_path = bucket_path[len('gs://'):]

    bucket = bucket_path.split('/')[0]
    filename = bucket_path[1 + len(bucket):]

    return upload_to_bucket(bucket, filename, json.dumps(obj), 'application/json', disable_cache)


def upload_to_bucket(bucket, blob_name, blob_contents, content_type='application/json', disable_cache=False):
    """
    Upload a blob of data to the specified google storage bucket.

    Args:
        bucket: str or google.cloud.storage.Bucket
            The bucket to upload the data to.
        blob_name: str
            The name of the blob to upload.
        blob_contents: str
            The data to upload.
    """
    if isinstance(bucket, str):
        from google.cloud import storage
        storage_client = storage.Client()
        bucket = storage_client.get_bucket(bucket)

    blob = bucket.blob(blob_name)
    if disable_cache:
        blob.cache_control = 'public, no-store'
    blob.upload_from_string(blob_contents, content_type)
    return blob.public_url


def make_bucket_public(bucket=None):
    """
    To host data in google storage for neuroglancer for general access,
    you need to make the bucket public and configure permissive CORS settings.

    This function does that.

    This feature is also installed along with ngsidekick as a command-line tool:

        .. code-block:: bash

            make-bucket-public gs://mybucket/mydir

    Args:
        bucket: str or None
            The bucket to make public, such as 'gs://mybucket/mydir'
            If None, then the bucket is read from the command line.
    """
    if bucket is None:
        parser = argparse.ArgumentParser()
        parser.add_argument('bucket')
        bucket = parser.parse_args().bucket
    if bucket.startswith('gs://'):
        bucket = bucket[len('gs://'):]

    subprocess.run(f'gsutil iam ch allUsers:objectViewer gs://{bucket}', shell=True, check=True)

    with tempfile.NamedTemporaryFile('w') as f:
        cors_settings = [{
            "maxAgeSeconds": 3600,
            "method": ["GET"],
            "origin": ["*"],
            "responseHeader": ["Content-Type", "Range"]
        }]
        json.dump(cors_settings, f)
        f.flush()
        subprocess.run(f'gsutil cors set {f.name} gs://{bucket}', shell=True, check=True)

    print(f"Configured bucket for public neuroglancer access: gs://{bucket}")



"""
Functions for encoding or parsing data for ``local://annotations`` sources.
"""
import copy
from textwrap import indent, dedent

import numpy as np
import pandas as pd

from ..state_utils import parse_nglink
from .util import annotation_property_specs

def extract_local_annotations(link, *, link_index=None, user=None, visible_only=False):
    """
    Extract local://annotations data from a neuroglancer link.
    The annotation coordinates (point, pointA, pointB, radii) are extracted
    into separate columns, named x,y,z/xa,ya,za/xb,yb,zb/rx,ry,rz
    (consistent with local_annotation_json() in this file).

    Args:
        link:
            Either a neuroglancer JSON state (dict), or a neuroglancer link with embedded state,
            or a neuroglancer link that references a JSON state file, such as
            https://neuroglancer-demo.appspot.com/#!gs://flyem-views/hemibrain/v1.2/base.json

        link_index:
            Deprecated.
            Adds a column named 'link_index' populated with the provided value in all rows.

        user:
            Deprecated.
            Adds a column named 'user' populated with the provided value in all rows.

        visible_only:
            If True, do not extract annotations from layers that are not currently
            visible in the given link state.

    Returns:
        DataFrame
    """
    if isinstance(link, str):
        link = parse_nglink(link)
    annotation_layers = [layer for layer in link['layers'] if layer['type'] == "annotation"]

    dfs = []
    for layer in annotation_layers:
        if visible_only and (layer.get('archived', False) or not layer.get('visible', True)):
            continue

        try:
            _df = pd.DataFrame(layer['annotations'])
        except KeyError as e:
            continue
        _df['layer'] = layer['name']
        dfs.append(_df)

    df = pd.concat(dfs, ignore_index=True)
    if 'point' in df.columns:
        rows = df['point'].notnull()
        df.loc[rows, [*'xyz']] = df.loc[rows, 'point'].tolist()
    if 'pointA' in df.columns:
        rows = df['pointA'].notnull()
        df.loc[rows, ['xa', 'ya', 'za']] = df.loc[rows, 'pointA'].tolist()
    if 'pointB' in df.columns:
        rows = df['pointB'].notnull()
        df.loc[rows, ['xb', 'yb', 'zb']] = df.loc[rows, 'pointB'].tolist()
    if 'radii' in df.columns:
        rows = df['radii'].notnull()
        df.loc[rows, ['rx', 'ry', 'rz']] = df.loc[rows, 'radii'].tolist()

    # Convert to int if possible.
    for col in [*'xyz', 'xa', 'ya', 'za', 'xb', 'yb', 'zb', 'rx', 'ry', 'rz', 'radii']:
        if col in df and df[col].notnull().all() and not (df[col] % 1).any():
            df[col] = df[col].astype(np.int64)

    if link_index is not None:
        df['link_index'] = link_index
    if user is not None:
        df['user'] = user

    df = df.drop(columns=['point', 'pointA', 'pointB', 'radii'], errors='ignore')
    cols = ['link_index', 'user', 'layer', 'type', *'xyz', 'rx', 'ry', 'rz', 'xa', 'ya', 'za', 'xb', 'yb', 'zb', 'id', 'description']
    df = df[[c for c in cols if c in df.columns]]
    return df


# Tip: Here's a nice repo with lots of colormaps implemented in GLSL.
# https://github.com/kbinani/colormap-shaders
SHADER_FMT = dedent("""\
    void main() {{
        setColor(defaultColor());
        setPointMarkerSize({size:.1f});
    }}
""")

LOCAL_ANNOTATION_JSON = {
    "name": "annotations",
    "type": "annotation",
    "source": {
        "url": "local://annotations",
        "transform": {
            "outputDimensions": {
                "x": [
                    8e-09,
                    "m"
                ],
                "y": [
                    8e-09,
                    "m"
                ],
                "z": [
                    8e-09,
                    "m"
                ]
            }
        }
    },
    "tool": "annotatePoint",
    "shader": "\nvoid main() {\n  setColor(defaultColor());\n  setPointMarkerSize(8.0);\n}\n",
    "panels": [
        {
            "row": 1,
            "flex": 1.22,
            "tab": "annotations"
        }
    ],
    "annotations": [
        # {
        #     "point": [23367, 35249, 68171],
        #     "type": "point",
        #     "id": "149909688276769607",
        #     "description": "soma"
        # },
    ]
}


def local_annotation_json(df, name="annotations", color="#ffff00", size=8.0, linkedSegmentationLayer=None,
                          show_panel=False, properties=[], shader=None, res_nm_xyz=(8,8,8)):
    """
    Construct the JSON data for a neuroglancer local annotations layer.
    This does not result in a complete neuroglancer link; it results in something
    that can be added to the layers list in the neuroglancer viewer JSON state.

    Note:
        This function's signature will likely change in the future.

    Args:
        df:
            DataFrame containing the annotation data.
            Which columns you must provide depends on which annotation type(s) you want to display.

            - For point annotations, provide ['x', 'y', 'z']
            - For line annotations or axis_aligned_bounding_box annotations,
              provide ['xa', 'ya', 'za', 'xb', 'yb', 'zb']
            - For ellipsoid annotations, provide ['x', 'y', 'z', 'rx', 'ry', 'rz']
              for the center point and radii.

            You may also provide a column 'type' to explicitly set the annotation type.
            In some cases, 'type' isn't needed since annotation type can be inferred from the
            columns you provided. But in the case of line and box annotations, the input
            columns are the same, so you must provide a 'type' column.

            You may also provide additional columns to use as annotation properties,
            in which case they should be listed in the 'properties' argument. (See below.)

        name:
            The name of the annotation layer

        color:
            The default color for annotations, which can be overridden by the annotation shader.

        size:
            The annotation size to hard-code into the default annotation shader used by this function.
            (Only used for points and line endpoints.)

        linkedSegmentationLayer:
            If the annotations should be associated with another layer in the view,
            this specifies the name of that layer.
            This function sets the 'filterBySegmentation' key to hide annotations from non-selected segments.
            If you are providing a linkedSegmentationLayer, your dataframe should contain
            a 'segments' column to indicate which segments are associated with each annotation.

        show_panel:
            If True, the selection panel will be visible in the side bar by default.

        properties:
            The list column names to use as annotation properties.
            Properties are visible in the selection panel when an annotation is selected,
            and they can also be used in annotation shaders via special functions neuroglancer
            defines for each property.  For example, for a property named 'confidence',
            you could write setPointMarkerSize(prop_confidence()) in your annotation shader.

            This function supports annotation color proprties via strings (e.g. '#ffffff') and
            also annotation 'enum' properties if you pass them via pandas categorical columns.

            By default, the annotation IDs are the same as the column names and the annotation types are inferred.
            You can override the property 'spec' by supplying a dict-of-dicts here instead of a list of columns:

            .. code-block:: python

                properties={
                    "my_column": {
                        "id": "my property",
                        "description": "This is my annotation property.",
                        "type": "float32",
                        "enum_values": [0.0, 0.5, 1.0],
                        "enum_labels": ["nothing", "something", "everything"],
                    },
                    "another_column: {...}
                }

    Returns:
        dict (JSON data)
    """
    df = _standardize_annotation_dataframe(df)
    if isinstance(properties, str):
        properties = [properties]

    data = copy.deepcopy(LOCAL_ANNOTATION_JSON)
    res_m = (np.array(res_nm_xyz) * 1e-9).tolist()
    output_dim = {k: [r, 'm'] for k,r in zip('xyz', res_m)}
    data['source']['transform']['outputDimensions'] = output_dim
    data['name'] = name
    data['annotationColor'] = color

    if shader:
        data['shader'] = shader
    else:
        data['shader'] = _default_shader(df['type'].unique(), size)

    if linkedSegmentationLayer:
        data['linkedSegmentationLayer'] = linkedSegmentationLayer
        data['filterBySegmentation'] = ['segments']

    if not show_panel:
        del data['panels']

    prop_specs = annotation_property_specs(df, properties)
    if prop_specs:
        data['annotationProperties'] = prop_specs
        properties = [p['id'] for p in prop_specs]
    
    data['annotations'].clear()
    data['annotations'] = _annotation_list_json(
        df, linkedSegmentationLayer, properties
    )
    return data


def _annotation_list_json(df, linkedSegmentationLayer, properties):
    """
    Helper for local_annotation_json().

    Generate the list of annotations for an annotation layer,
    assuming the input dataframe has already been pre-conditioned.
    """
    # Replace categoricals with their integer codes.
    # The corresponding enum_labels are already stored in the property specs
    for col in properties:
        if df[col].dtype == "category":
            df[col] = df[col].cat.codes

    annotations = []
    for row in df.itertuples():
        entry = {}
        entry['type'] = row.type
        entry['id'] = row.id
        if 'description' in df.columns:
            entry['description'] = row.description

        if row.type == 'point':
            entry['point'] = [row.x, row.y, row.z]
        elif row.type in ('line', 'axis_aligned_bounding_box'):
            entry['pointA'] = [row.xa, row.ya, row.za]
            entry['pointB'] = [row.xb, row.yb, row.zb]
        elif row.type == 'ellipsoid':
            entry['point'] = [row.x, row.y, row.z]
            entry['radii'] = [row.rx, row.ry, row.rz]
        else:
            raise RuntimeError(f'Invalid annotation type: {row.type}')

        if linkedSegmentationLayer and 'segments' in df.columns:
            segments = row.segments
            if not hasattr(segments, '__len__'):
                segments = [segments]
            segments = [str(int(s)) for s in segments]
            entry['segments'] = segments

        if properties:
            entry['props'] = [getattr(row, prop) for prop in properties]

        annotations.append(entry)
    return annotations


def _standardize_annotation_dataframe(df):
    """
    Helper for local_annotation_json().
    Add empty columns as needed until the dataframe
    has all possible annotation columns.

    Also populate the 'type' column with inferred
    annotation types based on the columns the user DID provide.
    """
    df = df.copy()
    id_cols = [*'xyz', 'xa', 'ya', 'za', 'xb', 'yb', 'zb', 'rx', 'ry', 'rz', 'type']
    for col in id_cols:
        if col not in df.columns:
            df[col] = np.nan

    if 'id' in df.columns:
        df['id'] = df['id'].astype(str)
    else:
        df['id'] = [
            str(hex(abs(hash(tuple(x)))))
            for x in df[id_cols].values.tolist()
        ]

    is_point_or_ellipsoid = df[[*'xyz']].notnull().all(axis=1)
    is_line_or_box = df[['xa', 'ya', 'za', 'xb', 'yb', 'zb']].notnull().all(axis=1)
    assert (is_point_or_ellipsoid ^ is_line_or_box).all(), \
        "You must supply either [x,y,z] or [xa,ya,za,xb,yb,zb] for every row (and not both)."

    df['type'] = df['type'].fillna(
        df['rx'].isnull().map({
            True: None,
            False: 'ellipsoid'
        })
    )

    # We have no way of choosing between 'line' and 'axis_aligned_bounding_box'
    # unless the user provides the 'type' explicitly.
    # We default to 'axis_aligned_bounding_box' because it's harder to type than 'line' :-)
    df['type'] = df['type'].fillna(
        df['x'].isnull().map({
            True: 'axis_aligned_bounding_box',
            False: 'point'
        })
    )
    return df


def _default_shader(annotation_types, default_size):
    """
    Create a default annotation shader that is pre-populated with
    the annotation API functions so you don't have to look them up.
    """
    shader_body = ""
    if 'point' in annotation_types:
        # Note:
        #   In older versions of neuroglancer,
        #   setPointMarkerColor(vec3) doesn't exist yet,
        #   so we must use vec4.
        #   https://github.com/google/neuroglancer/pull/475
        shader_body += dedent(f"""\
            //
            // Point Marker API
            //
            setPointMarkerSize({default_size});
            setPointMarkerColor(vec4(defaultColor(), 1.0));
            setPointMarkerBorderWidth(1.0);
            setPointMarkerBorderColor(defaultColor());
        """)

    if 'line' in annotation_types:
        shader_body += dedent(f"""\
            //
            // Line API
            //
            setLineColor(defaultColor());
            setEndpointMarkerSize({default_size}, {default_size});
            setEndpointMarkerColor(defaultColor(), defaultColor());
            setEndpointMarkerBorderWidth(1.0, 1.0);
            setEndpointMarkerBorderColor(defaultColor(), defaultColor());
        """)

    if 'axis_aligned_bounding_box' in annotation_types:
        shader_body += dedent("""\
            //
            // Bounding Box API
            //
            setBoundingBoxBorderWidth(1.0);
            setBoundingBoxBorderColor(defaultColor());
            setBoundingBoxFillColor(vec4(defaultColor(), 0.5));
        """)

    if 'ellipsoid' in annotation_types:
        shader_body += dedent("""\
            //
            // Ellipsoid API
            //
            setEllipsoidFillColor(defaultColor());
        """)

    shader_main = dedent(f"""\
        void main() {{
            {indent(shader_body, ' '*12)[12:]}\
        }}
    """)

    return shader_main



import pytest
import weakref
import numpy as np
import pandas as pd

from neuroglancer.coordinate_space import CoordinateSpace
from ngsidekick.annotations.precomputed import write_precomputed_annotations, TableHandle


@pytest.fixture(scope='session')
def test_output_dir(tmp_path_factory):
    """
    Create a common temporary directory for all tests to write to.
    This is session-scoped, so all tests share the same temp directory.
    """
    return tmp_path_factory.mktemp('test_annotations')


def _point_testdata(
    num_clusters=100,
    num_points_per_cluster=1_000,
    cluster_spacing=[10, 100, 1000],
    point_spacing=[1, 10, 100]
):
    centers = np.random.normal(0, cluster_spacing, (num_clusters, 3))
    points = []
    for center in centers:
        p = np.random.normal(center, point_spacing, (num_points_per_cluster, 3))
        points.append(p)
    points = np.concatenate(points, axis=0)
    points_df = pd.DataFrame(points, columns=[*'xyz'])
    points_df['cluster_id'] = np.repeat(np.arange(num_clusters), num_points_per_cluster).astype(np.uint32)

    from bokeh.palettes import Category10
    colormap = pd.Series(Category10[10])

    def hex_to_rgb(h):
        return [int(c, base=16) for c in (h[1:3], h[3:5], h[5:7])]

    rgb = colormap.loc[points_df['cluster_id'] % 10].map(hex_to_rgb).tolist()
    points_df[['cluster_color_r', 'cluster_color_g', 'cluster_color_b']] = rgb
    points_df['cluster_color_a'] = 255

    return points_df

@pytest.fixture
def point_testdata():
    return _point_testdata()

@pytest.fixture
def pointpair_testdata(
    num_clusters=100,
    num_points_per_cluster=1_000,
    cluster_spacing=[10, 100, 1000],
    point_spacing=[1, 10, 100],
    line_sigma=[.1, 1, 10]
):
    midpoints = _point_testdata(
        num_clusters=num_clusters,
        num_points_per_cluster=num_points_per_cluster,
        cluster_spacing=cluster_spacing,
        point_spacing=point_spacing
    )
    deltas = np.random.normal(0, line_sigma, (len(midpoints), 3))
    starts = midpoints[[*'xyz']] - deltas
    ends = midpoints[[*'xyz']] + deltas
    pairs = pd.DataFrame(
        np.concatenate([starts, ends], axis=1),
        columns=['xa', 'ya', 'za', 'xb', 'yb', 'zb']
    )
    return pd.concat((pairs, midpoints[[c for c in midpoints.columns if c not in 'xyz']]), axis=1)


@pytest.fixture
def ellipsoid_testdata(
    num_clusters=100,
    num_points_per_cluster=1_000,
    cluster_spacing=[10, 100, 1000],
    point_spacing=[1, 10, 100],
    radius_sigma=[0.2, 2, 20],
    radius_mean=[1, 10, 100]
):
    ellipsoid_centers = _point_testdata(
        num_clusters=num_clusters,
        num_points_per_cluster=num_points_per_cluster,
        cluster_spacing=cluster_spacing,
        point_spacing=point_spacing
    )
    # Create radii for each ellipsoid (one per row, not per cluster)
    num_ellipsoids = len(ellipsoid_centers)
    radii = np.abs(radius_mean + np.random.normal(0, radius_sigma, (num_ellipsoids, 3)))
    ellipsoid_centers[['rx', 'ry', 'rz']] = radii
    return ellipsoid_centers


def test_point_annotations(point_testdata, test_output_dir):
    cs = CoordinateSpace(names=[*'xyz'], units=['m', 'm', 'm'], scales=[100, 10, 1])
    write_precomputed_annotations(
        point_testdata,
        cs,
        'point',
        ['cluster_color'],
        ['cluster_id'],
        output_dir=test_output_dir / 'test-point-annotations',
        write_by_spatial_chunk=True,
        num_spatial_levels=6,
        target_chunk_limit=10_000
    )


def test_line_annotations(pointpair_testdata, test_output_dir):
    cs = CoordinateSpace(names=[*'xyz'], units=['m', 'm', 'm'], scales=[100, 10, 1])
    write_precomputed_annotations(
        pointpair_testdata,
        cs,
        'line',
        ['cluster_color'],
        ['cluster_id'],
        output_dir=test_output_dir / 'test-line-annotations',
        write_by_spatial_chunk=True,
        num_spatial_levels=6,
        target_chunk_limit=10
    )


def test_box_annotations(pointpair_testdata, test_output_dir):
    cs = dict(names=[*'xyz'], units=['m', 'm', 'm'], scales=[100, 10, 1])
    write_precomputed_annotations(
        pointpair_testdata,
        cs,
        'axis_aligned_bounding_box',
        ['cluster_color'],
        ['cluster_id'],
        output_dir=test_output_dir / 'test-box-annotations',
        write_by_spatial_chunk=True,
        num_spatial_levels=6,
        target_chunk_limit=10
    )


def test_ellipsoid_annotations(ellipsoid_testdata, test_output_dir):
    cs = CoordinateSpace(names=[*'xyz'], units=['m', 'm', 'm'], scales=[100, 10, 1])
    write_precomputed_annotations(
        ellipsoid_testdata,
        cs,
        'ellipsoid',
        ['cluster_color'],
        ['cluster_id'],
        output_dir=test_output_dir / 'test-ellipsoid-annotations',
        write_by_spatial_chunk=True,
        num_spatial_levels=6,
        target_chunk_limit=10
    )

def test_early_data_deletion(point_testdata, test_output_dir):
    """
    If we wrap the data in a TableHandle and delete our own reference to the data,
    then the original reference should be invalid by the time the annotations are written.
    We can verify this with a weakref.
    """
    point_testdata = point_testdata.sample(100).copy()

    ref = weakref.ref(point_testdata)
    lines_handle = TableHandle(point_testdata)
    del point_testdata

    cs = CoordinateSpace(names=[*'xyz'], units=['m', 'm', 'm'], scales=[100, 10, 1])
    write_precomputed_annotations(
        lines_handle,
        cs,
        'point',
        output_dir=test_output_dir / 'test-tablehandle-deletion',
        write_by_spatial_chunk=True,
        num_spatial_levels=6,
        target_chunk_limit=10_000
    )
    
    assert lines_handle.df is None
    assert ref() is None


@pytest.mark.manual
def test_inspect_test_results(test_output_dir, point_testdata, pointpair_testdata, ellipsoid_testdata):
    """
    Inspect the exported annotations from this test suite in neuroglancer.
    This test is marked as 'manual' and skipped by default since it requires manual inspection.
    
    - Produce a neuroglancer link populated with the exported
      annotations, and print it to the console.
    - Launch cors_webserver.py to host the temporary directory
    - Wait for the user to interrupt.
    
    To run this test: pytest -s -m manual tests/test_precomputed_annotations.py
    """
    import subprocess
    import json
    from pathlib import Path
    
    # First, ensure all test data is written
    print("\nWriting test annotations to", test_output_dir)
    test_point_annotations(point_testdata, test_output_dir)
    test_line_annotations(pointpair_testdata, test_output_dir)
    test_box_annotations(pointpair_testdata, test_output_dir)
    test_ellipsoid_annotations(ellipsoid_testdata, test_output_dir)
    
    # Start CORS webserver in background
    port = 9000
    bind_addr = "127.0.0.1"
    
    print(f"\nStarting CORS webserver on http://{bind_addr}:{port}")
    print(f"Serving directory: {test_output_dir}")
    
    cors_server_script = Path(__file__).parent / "cors_webserver.py"
    server_process = subprocess.Popen(
        [
            "python",
            str(cors_server_script),
            "--port", str(port),
            "--bind", bind_addr,
            "--directory", str(test_output_dir)
        ],
        # Don't capture output so we can see Flask logs
        stdout=None,
        stderr=None
    )
    
    # Give server time to start
    import time
    time.sleep(1)
    
    # Construct neuroglancer state with all annotation layers
    base_url = f"http://{bind_addr}:{port}"
    
    ng_state = {
        "dimensions": {"x": [1, "m"], "y": [1, "m"], "z": [1, "m"]},
        "position": [0, 0, 0],
        "crossSectionScale": 10,
        "projectionScale": 10_000,
        "showSlices": False,
        "layers": [
            {
                "type": "annotation",
                "source": f"precomputed://{base_url}/test-point-annotations",
                "shader": """\nvoid main() {\n  setColor(prop_cluster_color());\n}\n""",
                "name": "points",
                "annotations": []
            },
            {
                "type": "annotation",
                "source": f"precomputed://{base_url}/test-line-annotations",
                "shader": """\nvoid main() {\n  setColor(prop_cluster_color());\n}\n""",
                "name": "lines",
                "annotations": []
            },
            {
                "type": "annotation",
                "source": f"precomputed://{base_url}/test-box-annotations",
                "shader": """\nvoid main() {\n  setColor(prop_cluster_color());\n}\n""",
                "name": "boxes",
                "annotations": []
            },
            {
                "type": "annotation",
                "source": f"precomputed://{base_url}/test-ellipsoid-annotations",
                "shader": """\nvoid main() {\n  setColor(prop_cluster_color());\n}\n""",
                "name": "ellipsoids",
                "annotations": []
            }
        ],
        "layout": "4panel"
    }
    
    # Encode state as JSON fragment
    import urllib.parse
    state_json = json.dumps(ng_state)
    encoded_state = urllib.parse.quote(state_json)
    
    neuroglancer_url = f"https://neuroglancer-demo.appspot.com/#!{encoded_state}"
    
    print("\n" + "="*80)
    print("Neuroglancer URL:")
    print(neuroglancer_url)
    print("="*80)
    print("\nPress Ctrl+C to stop the server and exit...")
    print()
    
    try:
        # Wait for user interrupt
        server_process.wait()
    except KeyboardInterrupt:
        print("\nShutting down server...")
    finally:
        server_process.terminate()
        server_process.wait(timeout=5)


if __name__ == "__main__":
    pytest.main(['-s', 'test_precomputed_annotations.py'])



from .segmentprops import segment_properties_json, segment_properties_to_dataframe
from .select_segment_properties import select_segment_properties

__all__ = [
    'segment_properties_json',
    'segment_properties_to_dataframe',
    'select_segment_properties',
]



import re
from collections.abc import Mapping, Collection

import numpy as np


# Copied from neuroglancer/write_annotations.py
_PROPERTY_DTYPES: dict[str, tuple[tuple[str] | tuple[str, tuple[int, ...]], int]] = {
    "uint8": (("|u1",), 1),
    "uint16": (("<u2",), 2),
    "uint32": (("<u4",), 3),
    "int8": (("|i1",), 1),
    "int16": (("<i2",), 2),
    "int32": (("<i4",), 4),
    "float32": (("<f4",), 4),
    "rgb": (("|u1", (3,)), 1),
    "rgba": (("|u1", (4,)), 1),
}


def annotation_property_specs(df, properties):
    """
    Given an input dataframe for annotations and a list of columns
    from which to generate annotation properties, generate a
    JSON property type specification.

    The annotation property type is inferred from each column's dtype.
    Categorical pandas columns result in neuroglancer enum annotation properties.

    In addition to constructing the property specs, this function sorts the properties
    according to the rules required for precomputed annotations (for padding efficiency).

    For local annotation layers, the property specs should be written
    as the 'annotationProperties' part of the layer JSON state.
   
    For precomputed annotations, the property specs should be written as
    the "properties" section of the info JSON.

    Args:
        df:
            DataFrame.  The property columns will be inspected to infer
            the ultimate property types (numeric vs enum vs color).
        properties:
            list of column names from which to generate properties,
            or a dict-of-dicts containing pre-formulated property specs
            as descrbed in the docstring for local_annotation_json().
    Returns:
        JSON dict
        The order of the properties may not be the same as what you passed in.
        They are sorted according to the layout rules of neuroglancer's
        precomputed annotation format.
    """
    try:
        from neuroglancer.viewer_state import AnnotationPropertySpec
    except ImportError:
        class AnnotationPropertySpec:
            pass

    if isinstance(properties, Mapping):
        property_specs = {}
        for key, spec in properties.items():
            if isinstance(spec, AnnotationPropertySpec):
                assert spec.id == key
                property_specs[key] = spec.to_json()
            else:
                property_specs[key] = spec
        property_specs = properties
    else:
        assert isinstance(properties, Collection)
        property_specs = {}
        for spec in properties:
            if isinstance(spec, AnnotationPropertySpec):
                property_specs[spec.id] = spec.to_json()
            elif isinstance(spec, dict):
                property_specs[spec['id']] = spec
            elif isinstance(spec, str):
                property_specs[spec] = {}
            else:
                raise ValueError(f"Invalid property spec: {spec}")

    default_property_specs = {}
    for propname in property_specs:
        if propname in df.columns:
            default_property_specs[propname] = {
                'id': propname,
                'type': _proptype(df[propname]),
            }
        elif {f'{propname}_{channel}' for channel in 'rgba'} <= set(df.columns):
            default_property_specs[propname] = {
                'id': propname,
                'type': 'rgba',
            }
        elif {f'{propname}_{channel}' for channel in 'rgb'} <= set(df.columns):
            default_property_specs[propname] = {
                'id': propname,
                'type': 'rgb',
            }
        else:
            raise ValueError(f"Property '{propname}' not found in dataframe")

    for col in default_property_specs.keys():
        if col in df and df[col].dtype == "category":
            if df[col].isnull().any():
                raise ValueError(f"Column {col} can't be used for an enum property because it has null values")

            cats = df[col].cat.categories.tolist()
            default_property_specs[col]['enum_values'] = [*range(len(cats))]
            default_property_specs[col]['enum_labels'] = cats

    property_specs = [
        {**default_property_specs[col], **property_specs[col]}
        for col in property_specs
    ]

    # For precomputed annotations, the property specs
    # must be sorted in the following order:
    # 32-bit properties, 16-bit properties, 8-bit-properties (including rgb and rgba)
    property_specs.sort(key=lambda x: -_PROPERTY_DTYPES[x['type']][1])

    # Validate the property names according to the neuroglancer spec.
    for prop in property_specs:
        if not re.match(r'^[a-z][a-zA-Z0-9_]*$', prop['id']):
            raise ValueError(f"Invalid property name: {prop}")

    return property_specs


def _proptype(s):
    """
    Helper for _annotation_property_specs().
    Given a Series, determine the corresponding neuroglancer property type.

    Returns: str
        Either a numeric type (e.g. 'uint16') or a color type ('rgb' or 'rgba').
    """
    if s.dtype in (np.float64, np.int64, np.uint64):
        raise RuntimeError("neuroglancer doesn't support 64-bit property types.")
    if s.dtype in (np.uint8, np.int8, np.uint16, np.int16, np.uint32, np.int32, np.float32):
        return str(s.dtype)

    if s.dtype == 'category':
        num_cats = len(s.dtype.categories)
        if s.cat.codes.dtype == np.int8:
            # Old versions of neuroglancer had a bug for int8 property types,
            # so we store them as uint8.
            # https://github.com/google/neuroglancer/pull/830
            return 'uint8'

        # Pandas already stores categorical codes with a minimal-width dtype
        return str(np.dtype(s.cat.codes.dtype))

    if s.dtype != object:
        raise RuntimeError(f"Unsupported property dtype: {s.dtype} for column {s.name}")

    is_str = s.map(lambda x: isinstance(x, str)).all()
    is_color = is_str and s.str.startswith('#').all()
    if not is_color:
        msg = (
            f"Column {s.name}: I don't know what to do with object dtype that isn't rbg or rgba.\n"
            "If you want to create an enum property, then supply a pandas Categorical column."
        )
        raise RuntimeError(msg)
    if (s.map(len) == len("#rrggbb")).all():
        return 'rgb'
    if (s.map(len) == len("#rrggbbaa")).all():
        return 'rgba'
    raise RuntimeError("Not valid RGB or RGBA colors")



from dataclasses import dataclass

from numba import njit
import numpy as np
import pandas as pd


@dataclass
class TableHandle:
    """
    A wrapper for a pandas DataFrame that can be provided to transfer ownership
    of the DataFrame to ``write_precomputed_annotations()``, which will delete
    the handle's reference to the DataFrame as soon as possible to save RAM.

    Example:

    .. code-block:: python

        >>> handle = TableHandle(df)
        >>> del df  # Delete your own reference to the original data
        >>> write_precomputed_annotations(handle, 'xyz', 'point')
    """
    df: pd.DataFrame | None = None


def _encode_uint64_series(s, dtype='<u8'):
    """
    Encode a pandas Series (or Index) of N values
    into a numpy array of N buffers (bytes objects).
    """
    id_buf = s.to_numpy(dtype).tobytes()
    id_bufs = [
        id_buf[offset:(offset+8)]
        for offset in range(0, len(id_buf), 8)
    ]
    return np.array(id_bufs, dtype=object)


def _geometry_cols(coord_names, annotation_type):
    """
    Determine the list of column groups that express
    the geometry of annotations of the given type.
    Point annotations have only one group,
    but other annotation types have two.
    
    Examples:
    
        >>> _geometry_cols([*'xyz'], 'point')
        [['x', 'y', 'z']]

        >>> _geometry_cols([*'xyz'], 'ellipsoid')
        [['x', 'y', 'z'], ['rx', 'ry', 'rz']]

        >>> _geometry_cols([*'xyz'], 'line')
        [['xa', 'ya', 'za'], ['xb', 'yb', 'zb']]

        >>> _geometry_cols([*'xyz'], 'axis_aligned_bounding_box')
        [['xa', 'ya', 'za'], ['xb', 'yb', 'zb']]
    """
    if annotation_type == 'point':
        return [[c for c in coord_names]]

    if annotation_type == 'ellipsoid':
        return [
            [c for c in coord_names],
            [f'r{c}' for c in coord_names]
        ]

    if annotation_type in ('line', 'axis_aligned_bounding_box'):
        return [
            [f'{c}a' for c in coord_names],
            [f'{c}b' for c in coord_names]
        ]

    raise ValueError(f"Annotation type {annotation_type} not supported")


def ndindex_array(shape):
    """
    Like np.ndindex, but returns an array.
    
    Example:
    
        >>>  _ndindex_array([3,4])
        array([[0, 0],
               [0, 1],
               [0, 2],
               [0, 3],
               [1, 0],
               [1, 1],
               [1, 2],
               [1, 3],
               [2, 0],
               [2, 1],
               [2, 2],
               [2, 3]])
    """
    return _ndindex_array(np.asarray(shape))

@njit
def _ndindex_array(shape):
    total_indices = np.prod(shape)
    ndim = len(shape)
    
    indices = np.zeros((total_indices, ndim), dtype=np.uint64)
    
    for linear_idx in range(total_indices):
        temp_idx = linear_idx

        # Work backwards through dimensions
        for dim in range(ndim - 1, -1, -1):
            indices[linear_idx, dim] = temp_idx % shape[dim]
            temp_idx = temp_idx // shape[dim]
    
    return indices



import logging
from ._write_buffers import _write_buffers

logger = logging.getLogger(__name__)


def _write_annotations_by_id(df, output_dir, write_sharded):
    """
    Write the annotations to the "Annotation ID Index", a subdirectory of output_dir.

    Args:
        df:
            DataFrame with columns ['id_buf', 'ann_buf'].

        output_dir:
            Directory to write the annotations to.
            A single subdirectory named 'by_id' will be created in output_dir.

        write_sharded:
            Whether to write the annotations in sharded format.

    Returns:
        JSON metadata to be written under the 'by_id' key in the top-level 'info' file.
        Currently, this is always {"key": "by_id"}
    """
    if 'rel_buf' in df.columns:
        ann_bufs = df['ann_buf'] + df['rel_buf']
    else:
        ann_bufs = df['ann_buf']

    logger.info("Writing annotations to 'by_id' index")
    metadata = _write_buffers(ann_bufs, output_dir, "by_id", write_sharded)
    return metadata




import re
import string
from typing import List, Dict
from pandas.api.types import is_numeric_dtype

import numpy as np
import pandas as pd

from ngsidekick.segmentprops.segmentprops import segment_properties_to_dataframe, segment_properties_json


def select_segment_properties(
    info: dict,
    subset: List[str],
    scalar_expressions: Dict[str, str] = {},
    tag_expressions: Dict[str, str] = {}
) -> dict:
    """
    Given the JSON data from a neuroglancer segment properties 'info' file,
    select a subset of the properties and/or tags and return the reduced JSON info.

    You can also add *new* properties and/or tags by providing expressions
    (in pandas eval syntax) to be evaluated using the source data.

    In addition to expressions, you can provide a template string to be evaluated
    with Python string formatting.

    The variables available to the expressions are the columns extracted from the
    source data as produced via `segment_properties_to_dataframe()`.
    The tag list is parsed into separate columns according to their prefixes
    (assuming ':' delimits the prefix).

    The 'label' column is special: It is the one that will be displayed in the neuroglancer segment list.

    Args:
        info:
            The JSON data from a neuroglancer segment properties 'info' file.

        subset:
            The subset of properties or tag categories to select.
            Special values:
            - '_all': Select all properties and tags.
            - '_all_tags': Select all tag categories.
            - '_default': Select all properties and tags that have less than 1000 unique values.
            - '_default_tags': Select all tag categories that have less than 1000 unique values.

        scalar_expressions:
            Expressions to be evaluated using the source data to compute new properties.

        tag_expressions:
            Expressions to be evaluated using the source data to compute new tags.

    Returns:
        The reduced/enhanced JSON data from a neuroglancer segment properties 'info' file.
    

    Note:
        For the purpose of expression evaluation, we treat '~' and missing values (untagged) as "".

    Examples:

    .. code-block:: python

        new_info = select_segment_properties(info, subset=['superclass', 'type'])
        new_info = select_segment_properties(
            info,
            scalar_expressions={
                'size': 'length * width * height',
                'label': "{type} ({superclass})"
            },
        )
        new_info = select_segment_properties(
            info,
            tag_expressions={
                'is_left_hemilineage_01': 'hemilineage in ("01A", "01B") and soma_side == "L"'
            }
        )
    
    See also:
        The `ngsidekick-server <https://github.com/janelia-flyem/ngsidekick-server>`_
        project wraps this function in a web service to enable dynamic segment property
        selection directly in neuroglancer (via user-provided source URLs).
    """

    scalar_df, tags_df = segment_properties_to_dataframe(info, return_separate_tags=True)

    # It's possible for a scalar column to also be used in the tags.
    # We'll drop the scalar and just keep the one from the tags,
    # since the tag column takes less RAM.
    scalar_df = scalar_df[[c for c in scalar_df.columns if c not in tags_df.columns]]

    full_df = pd.concat((scalar_df, tags_df), axis=1)
    new_df = _select_segment_properties_from_dataframe(full_df, {**scalar_expressions, **tag_expressions})
    new_tag_cols = list(tag_expressions.keys())

    if '_all' in subset:
        subset += [
            c for c in full_df.columns
            if c not in subset
        ]
        subset.remove('_all')

    if '_all_tags' in subset:
        subset += [
            c for c in full_df.columns
            if c not in subset and not is_numeric_dtype(full_df[c].dtype)
        ]
        subset.remove('_all_tags')

    if '_default' in subset:
        counts = full_df.nunique()
        subset += [
            c for c in full_df.columns
            if c not in subset and counts[c] <= 1000
        ]
        subset.remove('_default')

    if '_default_tags' in subset:
        counts = full_df.nunique()
        subset += [
            c for c in full_df.columns
            if c not in subset and counts[c] <= 1000 and not is_numeric_dtype(full_df[c].dtype)
        ]
        subset.remove('_default_tags')

    if invalid_subset := set(subset) - (set(tags_df.columns) | set(scalar_df.columns)):
        raise ValueError(f"Invalid segment properties: {', '.join(invalid_subset)}")

    subset_tags_df = tags_df[[c for c in subset if c in tags_df.columns]]
    subset_scalar_df = scalar_df[[c for c in subset if c in scalar_df.columns]]
    combined_df = pd.concat((new_df, subset_scalar_df, subset_tags_df), axis=1)
    return segment_properties_json(combined_df, tag_cols=[*new_tag_cols, *[c for c in subset if c in tags_df.columns]])


def _select_segment_properties_from_dataframe(
    full_df: pd.DataFrame,
    expressions: Dict[str, str] = {},
) -> pd.DataFrame:

    for col in full_df.columns.tolist():
        if full_df[col].dtype in ("category", "object", "string"):
            full_df[col] = full_df[col].astype('string').fillna('')
    
    # In some datasets, the segment properties use '~' to indicate an empty field
    # instead of the empty string to avoid sorting empty types near the top of the ID list.
    # For the purposes of expression evaluation, we treat '~' and '' as None.
    full_df_with_empty = full_df.replace('~', '')

    new_df = full_df[[]].copy()
    for name, expr in expressions.items():
        if template_names := string_template_names(expr):
            # Only evaluate the expression for rows in which at least
            # one of the referenced template names is not None.
            # Other rows are not evaluated and get an empty string by default.
            # (Special case: For the 'label' column, it will be replaced with '~' later.)
            new_df[name] = ''
            if invalid_template_names := set(template_names) - set(full_df_with_empty.columns):
                raise ValueError(f"Invalid segment properties: {', '.join(invalid_template_names)}")
            valid_rows = (full_df_with_empty[template_names] != '').any(axis=1)

            # This list comprehension is faster than using df.apply() with a lambda.
            new_df.loc[valid_rows, name] = [
                expr.format(**dict(zip(template_names, row)), locals={}, globals={}).strip()
                for row in full_df_with_empty.loc[valid_rows, template_names].values
            ]
        elif re.match(r'^[a-zA-Z_][a-zA-Z0-9_]*$', expr):
            # String appears not to be a template nor an expression;
            # it must just be a column name.
            new_df[name] = full_df[expr]
        else:
            new_df[name] = full_df_with_empty.eval(
                expr,
                local_dict={},
                global_dict={},
                engine='python'
            )

        # We append the newly computed column to the input dataframe so it
        # can be referenced in subsequent expressions if the user wants to.
        full_df_with_empty[name] = new_df[name]

    # It's annoying to have the empty labels sorted at the top of our ID list,
    # so we replace the empty string with '~' here to shove them to the bottom.
    if 'label' in new_df.columns:
        new_df['label'] = new_df['label'].replace('', '~')

    return new_df


def string_template_names(template: str) -> bool:
    try:
        return [name for (_, name, *_) in string.Formatter().parse(template) if name]
    except ValueError:
        return []



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import os
import inspect

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'ngsidekick'
copyright = '2025, HHMI'
author = 'Stuart Berg'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',
    'sphinx.ext.linkcode',
    'sphinx.ext.intersphinx',
    'sphinx_autodoc_typehints',
    'myst_parser',
]

templates_path = ['_templates']
exclude_patterns = []

# Support both .rst and .md files
source_suffix = {
    '.rst': 'restructuredtext',
    '.md': 'markdown',
}

# MyST-Parser configuration
myst_enable_extensions = [
    "colon_fence",      # ::: fences for directives
    "deflist",          # Definition lists
    "fieldlist",        # Field lists
    "html_image",       # HTML images
    "linkify",          # Auto-convert URLs to links
    "replacements",     # Text replacements
    "smartquotes",      # Smart quotes
    "tasklist",         # Task lists with [ ] and [x]
]

# Autodoc settings
autodoc_member_order = 'bysource'
autodoc_typehints = 'description'

# Intersphinx mapping
intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
    'numpy': ('https://numpy.org/doc/stable/', None),
    'pandas': ('https://pandas.pydata.org/docs/', None),
}



# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_book_theme'
html_static_path = ['_static']


# -- linkcode extension configuration ----------------------------------------
# Generate links to GitHub source code

def linkcode_resolve(domain, info):
    """
    Determine the URL corresponding to a Python object.
    """
    if domain != 'py':
        return None
    
    modname = info['module']
    fullname = info['fullname']
    
    # Import the module
    try:
        import importlib
        module = importlib.import_module(modname)
    except Exception:
        return None
    
    # Get the object
    obj = module
    for part in fullname.split('.'):
        try:
            obj = getattr(obj, part)
        except AttributeError:
            return None
    
    # Unwrap decorators
    try:
        obj = inspect.unwrap(obj)
    except Exception:
        pass
    
    # Get the source file
    try:
        source_file = inspect.getsourcefile(obj)
        if source_file is None:
            return None
    except Exception:
        return None
    
    # Get line numbers
    try:
        source, lineno = inspect.getsourcelines(obj)
        linespec = f"#L{lineno}-L{lineno + len(source) - 1}"
    except Exception:
        linespec = ""
    
    # Convert absolute path to relative path within the repo
    # The source files are in src/ngsidekick/
    try:
        # Find the 'src' directory in the path
        parts = source_file.split(os.sep)
        if 'src' in parts:
            idx = parts.index('src')
            relpath = os.path.join(*parts[idx:])
        else:
            return None
    except Exception:
        return None
    
    # Construct GitHub URL
    github_url = f"https://github.com/janelia-flyem/ngsidekick/blob/main/{relpath}{linespec}"
    return github_url



import json
from itertools import chain
from collections.abc import Iterable, Mapping
from typing import Optional, List, Dict, Union, Literal

import numpy as np
import pandas as pd


def segment_properties_json(
    df: pd.DataFrame,
    label_col: Optional[str] = None,
    *,
    description_col: Optional[str] = None,
    string_cols: List[str] = [],
    number_cols: List[str] = [],
    tag_cols: List[str] = [],
    tag_prefix_mode: Literal['all', 'disambiguate', None] = 'all',
    sort_tags: bool = False,
    tag_descriptions: Dict[str, str] = {},
    col_descriptions: Dict[str, str] = {},
    drop_empty: bool = True,
    output_path: Optional[str] = None,
) -> Dict[str, Union[str, Dict]]:
    """
    Given a DataFrame, construct the JSON representation for neuroglancer segment properties
    (i.e. segment properties 'info' file) according to the neuroglancer spec:
    https://github.com/google/neuroglancer/blob/master/src/datasource/precomputed/segment_properties.md

    Args:
        df:
            DataFrame whose index contains segment IDs.
            For clarity of intent, we require that the Index be named 'segment' or 'body'.

            All columns will be converted to segment properties. If a column is not
            explicitly listed in the following arguments, its property type will be inferred
            from the column dtype (or column name, in the case of label/description).

            The column dtypes can be string, category, or number, but 64-bit values
            will be downcast to 32-bit. Boolean columns are only valid as tags.

            The order of the input columns determines the order in which properties will be written.
            Neuroglancer preserves that order when displaying columns of numeric properties.

        label_col:
            Name of the column to use for the 'label' property, which is shown in neuroglancer by default.
            If the column happens to be named 'label', then you need not provide it here.
            Or, if your dataframe contains only one column (and it's non-numeric), we assume it is the label.

        description_col:
            Name of the column to use as the 'description' property.
            If you name the column 'description', then you need not provide it here.

        string_cols:
            Names of the columns to represent as 'string' properties.
            (Usually unnecessary.  Non-numeric columns become string properties by default,
            unless they are named in tag_cols.)

        number_cols:
            Names of the columns to represent as 'number' properties.
            (Usually unnecessary.  Numeric columns become number properties by default.)

        tag_cols:
            Names of the columns from which to generate the (combined) 'tags' property.
            If you want a column to be used for both tags _and_ a different property,
            be sure to list it explicitly in both arguments.  For example:

            .. code-block:: python

                segment_properties_json(df, label_col='cell_class', tag_cols=['cell_class', ...])

            In addition to string/category inputs, columns with dtype bool can also be
            used for tags, in which case the False items are discarded and the True items
            are tagged with the column name.

        tag_prefix_mode:
            Either ``'all'`` or ``'disambiguate'`` or ``None``.

            - If ``'all'``, then all tags will be prefixed with the name of their source column,
              (e.g. ``status:Anchor``), other than those which came from boolean columns.

            - If ``'disambiguate'``, then only tags which appear in multiple columns
              be prefixed with the name of their source column.

            - If None, then no disambiguation is performed.

        sort_tags:
            If True, order tags alphabetically (after prefixes are applied, if any) so they
            are listed in sorted order in the neuroglancer UI.
            Otherwise, tags are listed according to the order given by 'tag_cols',
            with tags from earlier columns preceding tags from later columns,
            but tags within each column will be sorted with respect to each other.
            Note: For categorical input columns with ordered categories, the sort order
            is determined by the categorical's intrinsic ordering, not alphanumeric sorting.

        tag_descriptions:
            Optional. A dict of {tag: description} describing each tag value.
            Missing tags will be given a default description.

        col_descriptions:
            Optional. A dict of {column: description} describing each property
            (input column) other than 'tags'.

        drop_empty:
            If any IDs in the input have no non-empty (null or "") properties,
            then drop them from the output entirely so they don't show up in
            neuroglancer's default segment list.

        output_path:
            If provided, export the JSON to a file.

    Returns:
        JSON data (as a dict) which should be written into an 'info' file to
        host neuroglancer precomputed segment properties.

    Example:

        .. code-block:: ipython

            In [41]: print(df)
            Out[41]:
                        type              class       hemilineage  has_soma  PreSyn  PostSyn
            segment
            910719  IN13B018   intrinsic_neuron               13B      True     786     3164
            160503    PFNm/p            central          DM3_CX_p      True      40      177
            908822  IN03B079   intrinsic_neuron               03B      True     176      951
            10552      DNp29  descending_neuron  putative_primary      True    1567     4559
            43797     MeVP11  visual_projection               NaN      True     158      504
            232845     R1-R6            sensory               NaN     False      25       18
            547654     R1-R6            sensory               NaN     False      57       29
            803197  IN07B019   intrinsic_neuron               TBD      True     796     3615
            75597      KCg-m            central              MBp4      True     223      802
            165485   vDeltaK            central               NaN      True      36      282

            In [42]: info = segment_properties_json(df, 'type', tag_cols=['class', 'hemilineage', 'has_soma'])

            In [43]: # Note: The output below has been indented by hand for clarity.
                ...: print(json.dumps(info))
            {
                "@type": "neuroglancer_segment_properties",
                "inline": {
                    "ids": ["910719", "160503", "908822", "10552", "43797", "232845", "547654", "803197", "75597", "165485"],
                    "properties": [
                        {
                            "id": "type",
                            "type": "label",
                            "values": ["IN13B018", "PFNm/p", "IN03B079", "DNp29", "MeVP11", "R1-R6", "R1-R6", "IN07B019", "KCg-m", "vDeltaK"]
                        },
                        {
                            "id": "PreSyn",
                            "type": "number",
                            "data_type": "int32",
                            "values": [786, 40, 176, 1567, 158, 25, 57, 796, 223, 36]
                        },
                        {
                            "id": "PostSyn",
                            "type": "number",
                            "data_type": "int32",
                            "values": [3164, 177, 951, 4559, 504, 18, 29, 3615, 802, 282]
                        },
                        {
                            "id": "tags",
                            "type": "tags",
                            "tags": [
                                "class:central", "class:descending_neuron", "class:intrinsic_neuron",
                                "class:sensory", "class:visual_projection",
                                "has_soma",
                                "hemilineage:03B", "hemilineage:13B", "hemilineage:DM3_CX_p", "hemilineage:MBp4",
                                "hemilineage:TBD", "hemilineage:putative_primary"
                            ],
                            "values": [
                                [2, 5, 7],
                                [0, 5, 8],
                                [2, 5, 6],
                                [1, 5, 11],
                                [4, 5],
                                [3],
                                [3],
                                [2, 5, 10],
                                [0, 5, 9],
                                [0, 5]
                            ]
                        }
                    ]
                }
            }
    """
    df, string_cols, number_cols, tag_cols = _validate_args(
        df, label_col, description_col, string_cols, number_cols,
        tag_cols, tag_prefix_mode, tag_descriptions, col_descriptions
    )

    if drop_empty:
        df = _drop_empty_rows(df)

    # Determine the neuroglancer property type for each column (except tags).
    scalar_types = _scalar_property_types(
        df, label_col, description_col, string_cols, number_cols, tag_cols
    )

    # Construct JSON for all properties except tags.
    json_props = []
    for col, prop_type in scalar_types.items():
        jsn = _scalar_property_json(df[col], prop_type, col_descriptions.get(col))
        json_props.append(jsn)

    # Construct JSON for tags.
    if len(tag_cols):
        jsn = _tags_property_json(df, tag_cols, tag_prefix_mode, sort_tags, tag_descriptions)
        json_props.append(jsn)

    # Assemble the complete output.
    segment_ids = [str(idx) for idx in df.index]
    info = {
        '@type': 'neuroglancer_segment_properties',
        'inline': {
            'ids': segment_ids,
            'properties': json_props
        }
    }

    if output_path:
        with open(output_path, 'w') as f:
            json.dump(info, f)

    return info


def _validate_args(df, label_col, description_col, string_cols, number_cols,
                   tag_cols, tag_prefix_mode, tag_descriptions, col_descriptions):
    """
    Basic checks and convenience conversions where appropriate.
    (Series -> DataFrame, str -> list)
    """
    if isinstance(df, pd.Series):
        df = df.to_frame()

    if isinstance(string_cols, str):
        string_cols = [string_cols]
    if isinstance(number_cols, str):
        number_cols = [number_cols]
    if isinstance(tag_cols, str):
        tag_cols = [tag_cols]

    assert isinstance(df, pd.DataFrame)
    assert isinstance(string_cols, Iterable)
    assert isinstance(number_cols, Iterable)
    assert isinstance(tag_cols, Iterable)
    assert isinstance(tag_descriptions, Mapping)
    assert isinstance(col_descriptions, Mapping)

    assert df.index.name in ('body', 'segment'), \
        f"Index name must be 'body' or 'segment', not '{df.index.name}'"
    assert tag_prefix_mode in ('all', 'disambiguate', None), \
        f"Invalid tag_prefix_mode: {tag_prefix_mode}."
    assert not (dupes := df.columns.duplicated()).any(), \
        f"Duplicated column names: {df.columns[dupes].tolist()}"

    # Check for columns that were listed in multiple arguments (except tag_cols).
    listed_scalar_cols = [label_col, description_col, *string_cols, *number_cols]
    listed_scalar_cols = [*filter(None, listed_scalar_cols)]
    listed_scalar_cols = pd.Series(listed_scalar_cols)
    dupes = listed_scalar_cols.loc[listed_scalar_cols.duplicated()].unique()
    if dupes := sorted(dupes):
        raise RuntimeError(f"Some columns were included in multiple arguments: {dupes}")

    return df, string_cols, number_cols, tag_cols


def _drop_empty_rows(df):
    """
    Drop rows which contain only NaN, "", and False, and thus
    will have no non-empty value for any neuroglancer property.
    """
    if df.empty:
        return df

    # Note that the only columns which are permitted to
    # have dtype 'bool' are the boolean tag cols (if any).
    # Therefore, False is considered 'empty'.
    bool_cols = df.select_dtypes(include=bool).columns
    other_cols = df.columns.difference(bool_cols)

    valid_bool = valid_other = False
    if len(bool_cols) > 0:
        valid_bool = df[bool_cols].any(axis=1)
    if len(other_cols) > 0:
        not_null = df[other_cols].notnull()
        not_empty = df[other_cols] != ''
        valid_other = (not_null & not_empty).any(axis=1)

    return df.loc[valid_bool | valid_other]


def _scalar_property_types(df, label_col, description_col, string_cols, number_cols, tag_cols):
    """
    Determine the names and neuroglancer property types of all scalar (non-tag)
    properties that can be extracted from the given DataFrame.

    The property types will be determined from the user's explicitly provided lists
    when available; otherwise they're inferred from the column name/dtype.

    Here, 'scalar' includes all non-tag types: number, string, label, description.
    """
    # Start with the user-provided types and then auto-type the other columns below.
    # Although we won't include tag columns in the result, we temporarily store
    # them in prop_types to avoid auto-typing them.
    # Note:
    #   The user is permitted to list tag columns in the scalar columns too,
    #   in which case the scalar type takes precedence in this function.
    prop_types = {c: 'tags' for c in tag_cols}
    prop_types |= {c: 'string' for c in string_cols}
    prop_types |= {c: 'number' for c in number_cols}

    if label_col:
        prop_types[label_col] = 'label'
    if description_col:
        prop_types[description_col] = 'description'

    if unknown_cols := set(prop_types.keys()) - set(df.columns):
        raise RuntimeError(f"Columns not found: {unknown_cols}")

    # For convenience, if there's only one column, and it wasn't explicitly listed,
    # AND it's non-numeric, we presume it's meant to be the 'label' column.
    if (
        len(df.columns) == 1
        and not prop_types
        and (df.dtypes.iloc[0] == 'string' or not np.issubdtype(df.dtypes.iloc[0], np.number))
    ):
        prop_types = {df.columns[0]: 'label'}

    # Infer the types of unlisted columns from either the name or dtype
    for name, dtype in df.dtypes.items():
        if dtype in (bool, 'boolean') and prop_types.get(name) != 'tags':  # noqa: E721
            raise RuntimeError(f"Column '{name}': Boolean columns are only valid as tag_cols")
        elif prop_types.get(name) == 'number' and not np.issubdtype(dtype, np.number):
            raise RuntimeError(f"Column '{name}': Not valid as number_cols (dtype: {dtype})")
        elif name in prop_types:
            continue
        elif name == 'label':
            prop_types['label'] = 'label'
        elif name == 'description':
            prop_types['description'] = 'description'
        elif dtype in ("object", "string"):
            prop_types[name] = 'string'
        elif np.issubdtype(dtype, np.number):
            prop_types[name] = 'number'
        else:
            prop_types[name] = 'string'

    # Re-order to match original input columns.
    # Property order determines appearance order in neuroglancer,
    # at least for numeric properties.
    prop_types = {c: prop_types[c] for c in df.columns}

    # Return scalar properties only
    prop_types = {k:v for k,v in prop_types.items() if v != 'tags'}
    return prop_types


def _scalar_property_json(s, prop_type, description):
    """
    Constructs the JSON for any segment property
    other than the 'tags' property.
    """
    if prop_type == 'number':
        return _scalar_number_property_json(s, description)

    prop = {
        'id': s.name,
        'type': prop_type,
        'values': s.fillna("").astype(str).tolist()
    }

    if description:
        prop['description'] = description

    return prop


def _scalar_number_property_json(s, description):
    """
    Constructs the JSON for a 'number' property.
    """
    dtype_name = s.dtype.name

    if s.dtype in (np.int64, np.uint64):
        dtype_name = _select_int64_downcast(s)

    if np.issubdtype(s.dtype, np.floating):
        dtype_name = 'float32'
        if s.isnull().any():
            raise RuntimeError(f"Numeric column '{s.name}' contains NaN.")

    prop = {
        'id': s.name,
        'type': 'number',
        'data_type': dtype_name,
        'values': s.tolist()
    }

    if description:
        prop['description'] = description

    return prop


def _select_int64_downcast(s):
    """
    Return either "int32" or "uint32" if the given (u)int64 Series
    can be safely converted to one of those dtypes without overflow.
    Otherwise, raise an error.
    """
    for dtype32 in (np.int32, np.uint32):
        info32 = np.iinfo(dtype32)
        if s.min() >= info32.min and s.max() <= info32.max:
            return info32.dtype.name

    raise RuntimeError(
        f"Can't create a property for column: '{s.name}'. "
        "Neuroglancer doesn't support 64-bit integer properties, "
        "and your data exceeds the limits of (u)int32."
    )


def _tags_property_json(df, tag_cols, tag_prefix_mode, sort_tags, tag_descriptions):
    """
    Constructs the JSON for the 'tags' segment property.
    """
    tags_df = df[[]].copy()

    # Individually convert each column to Categorical
    # before we combine into a unified Categorical below.
    for c in tag_cols:
        tags_df[c] = _convert_to_categorical(df[c])

    _insert_tag_prefixes(tags_df, tag_prefix_mode, df.dtypes)

    # Unified tag list: deduplicate, but preserve order by default.
    tag_sets = (dtype.categories for dtype in tags_df.dtypes)
    unique_tags = pd.Series([*chain(*tag_sets)]).unique().tolist()
    if sort_tags:
        unique_tags = sorted(unique_tags)

    # Convert all columns to a single unified Categorical dtype.
    tags_df = tags_df.astype(pd.CategoricalDtype(categories=unique_tags))

    # Tags are represented in JSON as a list-of-lists of sorted codes.
    codes_df = pd.DataFrame({c: s.cat.codes for c, s in tags_df.items()})
    sorted_codes = np.sort(codes_df.to_numpy(), axis=1).tolist()

    # Drop nulls
    for row in sorted_codes:
        row[:] = [x for x in row if x != -1]

    prop = {
        'id': 'tags',
        'type': 'tags',
        'tags': unique_tags,
        'values': sorted_codes,
    }

    if tag_descriptions:
        prop['tag_descriptions'] = _tag_description_list(unique_tags, tag_descriptions)

    return prop


def _convert_to_categorical(s):
    """
    Convert the given Series to a Categorical Series suitable for tags.
    """
    if s.dtype == 'category':
        s = s.cat.remove_unused_categories()

    if s.dtype in (bool, 'boolean'):
        s = s.astype('category', copy=False)
        s = s.cat.rename_categories({True: s.name})
        if False in s.dtype.categories:
            s = s.cat.remove_categories([False])

    s = s.astype('category', copy=False)
    s = _replace_spaces(s)

    # We interpret empty string as null
    if '' in s.dtype.categories:
        s = s.cat.remove_categories([''])

    return s


def _replace_spaces(s):
    """
    Replace all spaces in the given Categorical Series with underscores.
    (Spaces are forbidden in neuroglancer tags.)
    """
    renames = {
        cat: str(cat).replace(' ', '_')
        for cat in s.dtype.categories
    }

    try:
        return s.cat.rename_categories(renames)
    except ValueError:
        # If both "foo bar_baz" and "foo_bar baz" exist in the
        # original data, they will both map to "foo_bar_baz", and
        # rename_categories() complains about the duplicate category.
        return s.str.replace(' ', '_').astype('category')


def _insert_tag_prefixes(df, tag_prefix_mode, orig_dtypes):
    """
    Insert prefixes onto tags in df according to the tag_prefix_mode (if any).
    The columns of df must already have Categorical dtypes.

    Tags in columns which were originally bool (before we converted them
    to Categorical) require no prefix, so we refer to orig_dtypes to skip
    those columns.

    Modifies df in-place.
    """
    if tag_prefix_mode is None:
        return

    if tag_prefix_mode == 'disambiguate':
        _disambiguate_tags(df)
        return

    assert tag_prefix_mode == 'all'
    for c, s in list(df.items()):
        if orig_dtypes[c] in ('bool', 'boolean'):
            # Boolean columns need no prefix;
            # the column name is the entire tag.
            continue

        # Prefix must not contain spaces (forbidden by neuroglancer)
        # or colons (because we use ':' as the separator).
        prefix = s.name.replace(' ', '_').replace(':', '_')
        df[c] = s.cat.rename_categories([
            f'{prefix}:{cat}'
            for cat in s.dtype.categories
        ])


def _disambiguate_tags(df):
    """
    Given a dataframe in which all columns are Categoricals,
    find category values that are common across multiple columns
    and prepend a prefix (the column name) to such values
    to make sure no category value is duplicated from one column
    to the next.

    Modifies df in-place.
    """
    col_tags = [
        (col, tag)
        for col in df.columns
        for tag in df[col].dtype.categories
    ]
    col_tags = pd.DataFrame(col_tags, columns=['col', 'tag'])

    # Drop everything except the duplicate tags.
    is_dup = col_tags.duplicated('tag', keep=False)
    col_tags = col_tags.loc[is_dup].copy()

    # Prefix must not contain spaces (forbidden by neuroglancer)
    # or colons (because we use ':' as the separator).
    prefixes = col_tags['col'].str.replace(r'[: ]', '_')
    col_tags['new_tag'] = prefixes + ':' + col_tags['tag']

    for col, old_new_df in col_tags.groupby('col')[['tag', 'new_tag']]:
        renames = dict(old_new_df.to_numpy())
        df[col] = df[col].cat.rename_categories(renames)


def _tag_description_list(unique_tags, tag_descriptions):
    """
    Given the list of all tags and a mapping of tags to descriptions,
    return the descriptions in the same order as unique_tags.
    """
    tag_descriptions = {
        str(k).replace(' ', '_'): v
        for k,v in tag_descriptions.items()
    }

    td = []
    for t in unique_tags:
        d = tag_descriptions.get(t, None)

        # If we didn't find it, try stripping the tag prefix (if any)
        if not d and ':' in t:
            t2 = t[1+t.index(':'):]
            d = tag_descriptions.get(t2, None)

        # If we still didn't find it, emit the tag itself.
        td.append(d or t)

    return td


def segment_properties_to_dataframe(js: dict, consolidate_tags_by_prefix: bool = True, return_separate_tags: bool = False):
    """
    Converts segment properties info JSON to DataFrame.

    Args:
        js:
            The JSON representation of the segment properties.

        consolidate_tags_by_prefix:
            If True, consolidate boolean tag columns into Categorical columns
            according to their prefix, assuming the data was written
            using according to the naming conventions followed
            by segment_properties_json(..., tag_prefix_mode="all").
            Otherwise, a separate boolean column is created for each
            unique tag -- potentially tons of columns (and RAM)!

        return_separate_tags:
            If True, return two dataframes, for scalar columns and tag columns.
            Otherwise, combine them into a single dataframe.

    Returns:
        If return_separate_tags=True, return (scalar_df, tags_df).
        By default, concatenate them across columns into a single dataframe.
    """
    segment_ids = [*map(int, js['inline']['ids'])]
    segment_ids = pd.Index(segment_ids, name='segment')

    all_props = js['inline']['properties']
    scalar_props = [prop for prop in all_props if prop['type'] != 'tags']
    tags_props =   [prop for prop in all_props if prop['type'] == 'tags']

    scalar_values = {
        prop['id']: prop['values']
        for prop in scalar_props
    }

    scalar_dtypes = {
        prop['id']: prop['data_type']
        for prop in scalar_props
        if 'data_type' in prop
    }

    scalar_df = pd.DataFrame(scalar_values, segment_ids).astype(scalar_dtypes)

    if tags_props:
        tags_df = _convert_tags_to_dataframe(segment_ids, tags_props, consolidate_tags_by_prefix)
    else:
        tags_df = scalar_df[[]]

    if return_separate_tags:
        return scalar_df, tags_df

    # It's possible for a scalar column to also be used in the tags.
    # We'll drop the scalar and just keep the one from the tags,
    # since the tag column takes less RAM.
    scalar_df = scalar_df[[c for c in scalar_df.columns if c not in tags_df.columns]]
    return pd.concat((scalar_df, tags_df), axis=1)


def _convert_tags_to_dataframe(segment_ids, tags_props, consolidate_tags_by_prefix):
    code_lists = tags_props[0]['values']
    unique_tags = tags_props[0]['tags']

    segment_count = len(code_lists)
    tag_count = len(unique_tags)

    # Flatten the lists of codes,
    # but keep track of the rows they came from.
    flat_tag_codes = pd.Series([*chain(*code_lists)])
    code_rows = np.repeat(
        range(segment_count),
        [*map(len, code_lists)]
    )

    if not consolidate_tags_by_prefix:
        # Return a boolean column for every unique tag.
        # Could be tons of columns!
        flags_matrix = np.zeros((segment_count, tag_count), bool)
        flags_matrix[code_rows, flat_tag_codes] = True
        return pd.DataFrame(flags_matrix, segment_ids, unique_tags)

    tag_prefixes = pd.Series([tag.split(':', 1)[0] for tag in unique_tags]).astype('category')
    flat_prefixes = flat_tag_codes.map(tag_prefixes)

    # Assemble the flattened codes into a matrix (row per segment, column per prefix).
    # Default is -1 (categorical code for a missing value).
    code_matrix = -1 * np.ones((segment_count, pd.Series(tag_prefixes).nunique()), np.int32)
    code_matrix[code_rows, flat_prefixes.cat.codes] = flat_tag_codes

    tag_categoricals = {}
    for prefix, code_column in zip(tag_prefixes.cat.categories, code_matrix.T):
        # Initialize as a categorical that corresponds to all tags combined,
        # then reduce to a categorical for just this column's values (without prefixes).
        tag_cat = pd.Categorical.from_codes(code_column, unique_tags)
        tag_cat = tag_cat.remove_unused_categories()
        tag_cat = tag_cat.rename_categories({
            tag: tag.split(':', 1)[1] if ':' in tag else tag
            for tag in tag_cat.dtype.categories
        })
        tag_categoricals[prefix] = tag_cat

    return pd.DataFrame(tag_categoricals, segment_ids)



import os
import shutil
from dataclasses import dataclass
from typing import Literal

from tqdm import tqdm
import tensorstore as ts

from ._util import _encode_uint64_series


def _write_buffers(buf_series, output_dir, subdir, write_sharded):
    """
    Write the buffers to the appropriate subdirectory of output_dir,
    in sharded or unsharded format.

    Args:
        buf_series:
            pd.Series of dtype=object, whose values are buffers (bytes objects).
            The index of the series provides the keys under which each item is stored.

        output_dir:
            str
            The directory into which the exported annotations will be written.

        subdir:
            str
            The subdirectory into which the buffers will be written.
            If it already exists, it will be deleted before we (re)create it.

        write_sharded:
            bool
            If True, write the buffers in sharded format.
            If False, write one file per item.

    Returns:
        JSON metadata for the written data, including the key (subdir)
        and sharding spec if applicable.
    """
    if os.path.exists(f"{output_dir}/{subdir}"):
        shutil.rmtree(f"{output_dir}/{subdir}")

    if write_sharded:
        return _write_buffers_sharded(buf_series, output_dir, subdir)
    else:
        return _write_buffers_unsharded(buf_series, output_dir, subdir)


def _write_buffers_unsharded(buf_series, output_dir, subdir):
    """
    Write the buffers to the appropriate subdirectory of output_dir,
    in unsharded format, i.e. one file per item.

    The index of buf_series is used as the key for each item, after being
    converted to a string (as decimal values in the case of integer keys).

    Returns:
        JSON metadata, always {"key": subdir}
    """
    # In the unsharded format, the keys are just strings (e.g. decimal IDs).
    string_keys = buf_series.index.astype(str)
    buf_series = buf_series.set_axis(string_keys)

    # Since we're writing unsharded files, we could have just used
    # standard Python open() and write() here for each key.
    # Using tensorstore here is mostly just a matter of taste, but it will
    # become useful if we ever support alternative storage backends such as gcs.
    kvstore = ts.KvStore.open(f"file://{output_dir}/{subdir}/").result()

    # Using a transaction here is not necessary, at least for plain files.
    # I'm not sure if it helps or hurts, but it probably doesn't matter much
    # for small datasets, which is presumably what we're dealing with if the
    # user has chosen the unsharded format.
    with ts.Transaction() as txn:
        for segment_key, buf in tqdm(buf_series.items(), total=len(buf_series)):
            kvstore.with_transaction(txn)[segment_key] = buf

    metadata = {"key": subdir}
    return metadata


def _write_buffers_sharded(buf_series, output_dir, subdir):
    """
    Write the buffers to the appropriate subdirectory of output_dir,
    in sharded format.

    The index of buf_series is used as the key for each item,
    after being encoded as a bigendian uint64.

    Returns:
        JSON metadata, including the output "key" (subdir) and sharding spec.
    """
    # When writing sharded data, we must use encoded bigendian uint64 as the key.
    # https://github.com/google/neuroglancer/pull/522#issuecomment-1923137085
    bigendian_keys = _encode_uint64_series(buf_series.index, '>u8')
    buf_series = buf_series.set_axis(bigendian_keys)

    shard_spec = _choose_output_spec(
        total_count=len(buf_series),
        total_bytes=buf_series.map(len).sum(),  # fixme, might be slow
        hashtype='murmurhash3_x86_128',
        gzip_compress=True
    )
    spec = {
        "driver": "neuroglancer_uint64_sharded",
        "metadata": shard_spec.to_json(),
        "base": f"file://{output_dir}/{subdir}",
    }
    kvstore = ts.KvStore.open(spec).result()

    # Note:
    #   At the time of this writing, tensorstore uses a
    #   surprising amount of RAM to perform these writes.
    with ts.Transaction() as txn:
        for segment_key, buf in tqdm(buf_series.items(), total=len(buf_series)):
            kvstore.with_transaction(txn)[segment_key] = buf

    metadata = {
        "key": subdir,
        "sharding": shard_spec.to_json()
    }
    return metadata


@dataclass
class ShardSpec:
    """
    Copied from Forrest Collman's PR:
    https://github.com/google/neuroglancer/pull/522
    """
    type: str
    hash: Literal["murmurhash3_x86_128", "identity_hash"]
    preshift_bits: int
    shard_bits: int
    minishard_bits: int
    data_encoding: Literal["raw", "gzip"]
    minishard_index_encoding: Literal["raw", "gzip"]

    def to_json(self):
        return {
            "@type": self.type,
            "hash": self.hash,
            "preshift_bits": self.preshift_bits,
            "shard_bits": self.shard_bits,
            "minishard_bits": self.minishard_bits,
            "data_encoding": str(self.data_encoding),
            "minishard_index_encoding": str(self.minishard_index_encoding),
        }


def _choose_output_spec(
    total_count,
    total_bytes,
    hashtype: Literal["murmurhash3_x86_128", "identity_hash"] = "murmurhash3_x86_128",
    gzip_compress=True,
):
    """
    Copied from Forrest Collman's PR:
    https://github.com/google/neuroglancer/pull/522
    """
    import tensorstore as ts
    MINISHARD_TARGET_COUNT = 1000
    SHARD_TARGET_SIZE = 50000000

    # if total_count == 1:
    #     return None
    # if ts is None:
    #     return None

    # test if hashtype is valid
    if hashtype not in ["murmurhash3_x86_128", "identity_hash"]:
        raise ValueError(
            f"Invalid hashtype {hashtype}."
            "Must be one of 'murmurhash3_x86_128' "
            "or 'identity_hash'"
        )

    total_minishard_bits = 0
    while (total_count >> total_minishard_bits) > MINISHARD_TARGET_COUNT:
        total_minishard_bits += 1

    shard_bits = 0
    while (total_bytes >> shard_bits) > SHARD_TARGET_SIZE:
        shard_bits += 1

    preshift_bits = 0
    while MINISHARD_TARGET_COUNT >> preshift_bits:
        preshift_bits += 1

    minishard_bits = total_minishard_bits - min(total_minishard_bits, shard_bits)
    data_encoding: Literal["raw", "gzip"] = "raw"
    minishard_index_encoding: Literal["raw", "gzip"] = "raw"

    if gzip_compress:
        data_encoding = "gzip"
        minishard_index_encoding = "gzip"

    return ShardSpec(
        type="neuroglancer_uint64_sharded_v1",
        hash=hashtype,
        preshift_bits=preshift_bits,
        shard_bits=shard_bits,
        minishard_bits=minishard_bits,
        data_encoding=data_encoding,
        minishard_index_encoding=minishard_index_encoding,
    )
